#!/usr/bin/env python3

from typing import Optional
import torch as tt
import torch.nn.functional as F
from numpy.typing import NDArray
from typing import List
from rex_xai._utils import ff


class Prediction:
    def __init__(
        self,
        pred=None,
        conf=None,
        box=None,
        target=None,
        target_confidence=None,
    ) -> None:
        self.classification: Optional[int] = pred
        self.confidence: Optional[float] = conf
        self.bounding_box: Optional[NDArray] = box
        self.target: Optional[int] = None if target is None else target.classification
        self.target_confidence: Optional[float] = target_confidence

    def __repr__(self) -> str:
        if self.bounding_box is None:
            if self.is_passing():
                return (
                    f"FOUND_CLASS: {self.classification}, CONF: {self.confidence:.5f}"
                )
            else:
                if self.target is None:
                    return f"FOUND_CLASS: {self.classification}, FOUND_CONF: {self.confidence:.5f}, TARGET_CLASS: n/a, TARGET_CONFIDENCE: n/a"
                else:
                    return f"FOUND_CLASS: {self.classification}, FOUND_CONF: {self.confidence:.5f}, TARGET_CLASS: {self.target}, TARGET_CONFIDENCE: {ff(self.target_confidence, '.5f')}"

        return f"CLASS: {self.classification}, CONF: {self.confidence:.5f}, TARGET_CLASS: {self.target}, TARGET_CONFIDENCE: {ff(self.target_confidence, '.5f')}, BOUNDING_BOX: {self.bounding_box}"

    def get_class(self):
        return self.classification

    def is_empty(self):
        return self.classification is None or self.confidence is None

    def is_passing(self):
        return self.target == self.classification


def from_pytorch_tensor(tensor, target=None, binary_threshold=None) -> List[Prediction]:
    # TODO get this to handle binary models
    softmax_tensor = F.softmax(tensor, dim=1)
    prediction_scores, pred_labels = tt.topk(softmax_tensor, 1)
    predictions = []
    for i, (ps, pl) in enumerate(zip(prediction_scores, pred_labels)):
        p = Prediction(pl.item(), ps.item())
        if target is not None:
            p.target = target
            p.target_confidence = softmax_tensor[i, target.classification].item()
        predictions.append(p)

    return predictions



#!/usr/bin/env python
from datetime import datetime
import zlib
import torch as tt
import sqlalchemy as sa
from sqlalchemy import Boolean, Float, String, create_engine
from sqlalchemy import Column, Integer, Unicode
from sqlalchemy.orm import sessionmaker, DeclarativeBase
from ast import literal_eval

import pandas as pd
import numpy as np

from rex_xai.logger import logger
from rex_xai.config import CausalArgs, Strategy
from rex_xai.extraction import Explanation
from rex_xai.multi_explanation import MultiExplanation


def _dataframe(db, table):
    return pd.read_sql_table(table, f"sqlite:///{db}")


def _to_numpy(buffer, shape, dtype):
    return np.frombuffer(zlib.decompress(buffer), dtype=dtype).reshape(shape)


def db_to_pandas(db, dtype=np.float32, table="rex", process=True):
    """for interactive use"""
    df = _dataframe(db, table=table)

    if process:
        df["responsibility"] = df.apply(
            lambda row: _to_numpy(
                row["responsibility"], literal_eval(row["responsibility_shape"]), dtype
            ),
            axis=1,
        )
        #
        df["explanation"] = df.apply(
            lambda row: _to_numpy(
                row["explanation"], literal_eval(row["explanation_shape"]), np.bool_
            ),
            axis=1,
        )

    return df


def update_database(
    db,
    explanation: Explanation | MultiExplanation,  # type: ignore
    time_taken=None,
    multi=False,
):
    target_map = explanation.target_map

    if isinstance(target_map, tt.Tensor):
        target_map = target_map.detach().cpu().numpy()

    target = explanation.data.target
    if target is None:
        logger.warning("unable to update database as target is None")
        return
    classification = int(target.classification)  # type: ignore

    if not multi:
        final_mask = explanation.final_mask
        if explanation.final_mask is None:
            logger.warning("unable to update database as explanation is empty")
            return
        if isinstance(final_mask, tt.Tensor):
            final_mask = final_mask.detach().cpu().numpy()

        explanation_confidence = explanation.explanation_confidence

        add_to_database(
            db,
            explanation.args,
            classification,
            target.confidence,
            target_map,
            final_mask,
            explanation_confidence,
            time_taken,
            explanation.run_stats["total_passing"],
            explanation.run_stats["total_failing"],
            explanation.run_stats["max_depth_reached"],
            explanation.run_stats["avg_box_size"],
        )

    else:
        if type(explanation) is not MultiExplanation:
            logger.warning(
                "unable to update database, multi=True is only valid for MultiExplanation objects"
            )
            return
        else:
            for c, final_mask in enumerate(explanation.explanations):
                if isinstance(final_mask, tt.Tensor):
                    final_mask = final_mask.detach().cpu().numpy()
                add_to_database(
                    db,
                    explanation.args,
                    classification,
                    target.confidence,
                    target_map,
                    final_mask,
                    explanation.explanation_confidences[c],
                    time_taken,
                    explanation.run_stats["total_passing"],
                    explanation.run_stats["total_failing"],
                    explanation.run_stats["max_depth_reached"],
                    explanation.run_stats["avg_box_size"],
                    multi=multi,
                    multi_no=c,
                )


def add_to_database(
    db,
    args: CausalArgs,
    target,
    confidence,
    responsibility,
    explanation,
    explanation_confidence,
    time_taken,
    passing,
    failing,
    depth_reached,
    avg_box_size,
    multi=False,
    multi_no=None,
):
    if multi:
        id = hash(str(datetime.now().time()) + str(multi_no))
    else:
        id = hash(str(datetime.now().time()))

    responsibility_shape = responsibility.shape
    explanation_shape = explanation.shape

    object = DataBaseEntry(
        id,
        args.path,
        target,
        confidence,
        responsibility,
        responsibility_shape,
        explanation,
        explanation_shape,
        explanation_confidence,
        time_taken,
        depth_reached=depth_reached,
        avg_box_size=avg_box_size,
        tree_depth=args.tree_depth,
        search_limit=args.search_limit,
        iters=args.iters,
        min_size=args.min_box_size,
        distribution=str(args.distribution),
        distribution_args=str(args.distribution_args),
    )
    # if object is not None:
    object.multi = multi
    object.multi_no = multi_no
    object.passing = passing
    object.failing = failing
    object.total_work = passing + failing
    object.method = str(args.strategy)
    if args.strategy == Strategy.Spatial:
        object.spatial_radius = args.spatial_initial_radius
        object.spatial_eta = args.spatial_radius_eta
    if args.strategy == Strategy.MultiSpotlight:
        object.spotlights = args.spotlights
        object.spotlight_size = args.spotlight_size
        object.spotlight_eta = args.spotlight_eta
        object.obj_function = args.spotlight_objective_function

    db.add(object)
    db.commit()


class Base(DeclarativeBase):
    pass


class NumpyType(sa.types.TypeDecorator):
    impl = sa.types.LargeBinary

    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is not None:
            return zlib.compress(value, 9)

    def process_result_value(self, value, dialect):
        return value


class DataBaseEntry(Base):
    __tablename__ = "rex"
    id = Column(Integer, primary_key=True)
    path = Column(Unicode(100))
    target = Column(Integer)
    confidence = Column(Float)
    time = Column(Float)
    responsibility = Column(NumpyType)
    responsibility_shape = Column(Unicode)
    total_work = Column(Integer)
    passing = Column(Integer)
    failing = Column(Integer)
    explanation = Column(NumpyType)
    explanation_shape = Column(Unicode)
    explanation_confidence = Column(Float)
    multi = Column(Boolean)
    multi_no = Column(Integer)

    # causal specific columns
    depth_reached = Column(Integer)
    avg_box_size = Column(Float)
    tree_depth = Column(Integer)
    search_limit_per_iter = Column(Integer)
    iters = Column(Integer)
    min_size = Column(Integer)
    distribution = Column(String)
    distribution_args = Column(String)

    # explanation specific columns
    spatial_radius = Column(Integer)
    spatial_eta = Column(Float)

    # spotlight columns
    method = Column(String)
    spotlights = Column(Integer)
    spotlight_size = Column(Integer)
    spotlight_eta = Column(Float)
    obj_function = Column(String)

    def __init__(
        self,
        id,
        path,
        target,
        confidence,
        responsibility,
        responsibility_shape,
        explanation,
        explanation_shape,
        explanation_confidence,
        time_taken,
        passing=None,
        failing=None,
        total_work=None,
        multi=False,
        multi_no=None,
        depth_reached=None,
        avg_box_size=None,
        tree_depth=None,
        search_limit=None,
        iters=None,
        min_size=None,
        distribution=None,
        distribution_args=None,
        initial_radius=None,
        radius_eta=None,
        method=None,
        spotlights=0,
        spotlight_size=0,
        spotlight_eta=0.0,
        obj_function=None,
    ):
        self.id = id
        self.path = path
        self.target = target
        self.confidence = confidence
        self.responsibility = responsibility
        self.responsibility_shape = str(responsibility_shape)
        self.explanation = explanation
        self.explanation_shape = str(explanation_shape)
        self.explanation_confidence = explanation_confidence
        self.time = time_taken
        self.total_work = total_work
        self.passing = passing
        self.failing = failing
        # multi status
        self.multi = multi
        self.multi_no = multi_no
        # causal
        self.depth_reached = depth_reached
        self.avg_box_size = avg_box_size
        self.tree_depth = tree_depth
        self.search_limit = search_limit
        self.iters = iters
        self.min_size = min_size
        self.distribution = distribution
        self.distribution_args = distribution_args
        # spatial
        self.spatial_radius = initial_radius
        self.spatial_eta = radius_eta
        self.method = method
        # spotlights
        self.spotlights = spotlights
        self.spotlight_size = spotlight_size
        self.spotlight_eta = spotlight_eta
        self.obj_function = obj_function


def initialise_rex_db(name, echo=False):
    engine = create_engine(f"sqlite:///{name}", echo=echo)
    Base.metadata.create_all(engine, tables=[DataBaseEntry.__table__], checkfirst=True)  # type: ignore
    Session = sessionmaker(bind=engine)
    s = Session()
    return s



#!/usr/bin/env python

"""generate multiple explanations from a responsibility landscape <pixel_ranking>"""

import os
import re
import numpy as np

import torch as tt
from itertools import combinations

from rex_xai.extraction import Explanation
from rex_xai.distributions import random_coords, Distribution
from rex_xai.logger import logger
from rex_xai._utils import powerset, clause_area, SpatialSearch
from rex_xai.visualisation import save_multi_explanation, save_image, plot_image_grid


class MultiExplanation(Explanation):
    def __init__(self, maps, prediction_func, data, args, run_stats):
        super().__init__(maps, prediction_func, data, args, run_stats)
        self.explanations = []
        self.explanation_confidences = []

    def __repr__(self) -> str:
        pred_func = repr(self.prediction_func)
        match_func_name = re.search(r"(<function .+) at", pred_func)
        if match_func_name:
            pred_func = match_func_name.group(1) + " >"

        run_stats = {k: round(v, 5) for k, v in self.run_stats.items()}

        exp_text = (
            "MultiExplanation:"
            + f"\n\tCausalArgs: {type(self.args)}"
            + f"\n\tData: {self.data}"
            + f"\n\tprediction function: {pred_func}"
            + f"\n\tResponsibilityMaps: {self.maps}"
            + f"\n\trun statistics: {run_stats} (5 dp)"
        )

        if len(self.explanations) == 0:
            return (
                exp_text
                + f"\n\texplanations: {self.explanations}"
                + f"\n\texplanation confidences: {self.explanation_confidences}"
            )
        else:
            return (
                exp_text
                + f"\n\texplanations: {len(self.explanations)} explanations of {type(self.explanations[0])} and shape {self.explanations[0].shape}"
                + f"\n\texplanation confidences: {[round(x, ndigits=5) for x in self.explanation_confidences]} (5 dp)"
            )

    def save(self, path, mask=None, multi=None, multi_style=None, clauses=None):
        if multi_style is None:
            multi_style = self.args.multi_style
        if multi_style == "contrastive":
            super().save(path, mask=self.final_mask)
        if multi_style == "separate":
            logger.info("saving explanations in multiple different files")
            for i, mask in enumerate(self.explanations):
                name, ext = os.path.splitext(path)
                exp_path = f"{name}_{i}{ext}"
                super().save(exp_path, mask=mask)
        elif multi_style == "composite":
            logger.info("using composite style to save explanations")
            if clauses is None:
                clause = range(0, len(self.explanations))
                save_multi_explanation(
                    self.explanations, self.data, self.args, clause=clause, path=path
                )
            else:
                for clause in clauses:
                    name, ext = os.path.splitext(path)
                    new_name = f"{name}_{clause}{ext}"
                    save_multi_explanation(
                        self.explanations,
                        self.data,
                        self.args,
                        clause=clause,
                        path=new_name,
                    )

    def show(self, path=None, multi_style=None, clauses=None):
        if multi_style is None:
            multi_style = self.args.multi_style
        outs = []

        for i, mask in enumerate(self.explanations):
            out = save_image(mask, self.data, self.args, path=None)
            outs.append(out)

        if multi_style == "separate":
            for i, mask in enumerate(self.explanations):
                out = save_image(mask, self.data, self.args, path=None)
                outs.append(out)

        elif multi_style == "composite":
            if clauses is None:
                clause = tuple([i for i in range(len(self.explanations))])
                out = save_multi_explanation(
                    self.explanations, self.data, self.args, clause=clause, path=None
                )
                outs.append(out)
            else:
                for clause in clauses:
                    out = save_multi_explanation(
                        self.explanations,
                        self.data,
                        self.args,
                        clause=clause,
                        path=None,
                    )
                    outs.append(out)

        if len(outs) > 1:
            plot_image_grid(outs)
        else:
            return outs[0]

    def extract(self, method=None):
        self.blank()
        # we start with the global max explanation
        logger.info("spotlight number 1 (global max)")
        conf = self._Explanation__global()  # type: ignore
        if self.final_mask is not None:
            self.explanations.append(self.final_mask)
            self.explanation_confidences.append(conf)
        self.blank()

        for i in range(0, self.args.spotlights - 1):
            logger.info("spotlight number %d", i + 2)
            conf = self.spotlight_search()
            if self.final_mask is not None:
                self.explanations.append(self.final_mask)
                self.explanation_confidences.append(conf)
            self.blank()
        logger.info(
            "ReX has found a total of %d explanations via spotlight search",
            len(self.explanations),
        )

    def __dice(self, d1, d2):
        """calculates dice coefficient between two numpy arrays of the same dimensions"""
        d_sum = d1.sum() + d2.sum()
        if d_sum == 0:
            return 0
        intersection = tt.logical_and(d1, d2)
        return np.abs((2.0 * intersection.sum() / d_sum).item())

    def separate_by(self, dice_coefficient: float, reverse=True):
        exps = []
        sizes = dict()

        for i, exp in enumerate(self.explanations):
            size = tt.count_nonzero(exp)
            if size > 0:
                exps.append(i)
                sizes[i] = size

        clause_len = 0
        clauses = []

        perms = combinations(exps, 2)
        bad_pairs = set()
        for perm in perms:
            left, right = perm
            if (
                self.__dice(self.explanations[left], self.explanations[right])
                > dice_coefficient
            ):
                bad_pairs.add(perm)

        for s in powerset(exps, reverse=reverse):
            found = True
            for bp in bad_pairs:
                if bp[0] in s and bp[1] in s:
                    found = False
                    break
            if found:
                if len(s) >= clause_len:
                    clause_len = len(s)
                    clauses.append(s)
                else:
                    break

        clauses = sorted(clauses, key=lambda x: clause_area(x, sizes))
        return clauses

    def contrastive(self, clauses):
        for clause in clauses:
            for subset in powerset(clause, reverse=False):
                mask = sum([self.explanations[x] for x in subset])
                mask = mask.to(tt.bool)  # type: ignore
                sufficient = tt.where(mask, self.data.data, self.data.mask_value)  # type: ignore
                counterfactual = tt.where(mask, self.data.mask_value, self.data.data)  # type: ignore
                ps = self.prediction_func(sufficient)[0]
                pn = self.prediction_func(counterfactual)[0]

                if (
                    ps.classification == self.data.target.classification  # type: ignore
                    and pn.classification != self.data.target.classification  # type: ignore
                ):
                    logger.info(
                        "found sufficient and necessary explanation of class %d, %d with confidence %f",
                        ps.classification,
                        pn.classification,
                        pn.confidence,
                    )
                    self.final_mask = mask
                    return subset
        logger.warn(
            "ReX is unable to find a counterfactual, so not producing an output. Exiting here..."
        )
        exit()

    def __random_step_from(self, origin, width, height, step=5):
        c, r = origin
        # flip a coin to move left (0) or right (1)
        c_dir = np.random.randint(0, 2)
        c = c - step if c_dir == 0 else c + step
        if c < 0:
            c = 0
        if c > width:
            c = width

        # flip a coin to move down (0) or up (1)
        r_dir = np.random.randint(0, 2)
        r = r - step if r_dir == 0 else r + step
        if r < 0:
            r = 0
        if r > height:
            r = height
        logger.debug(f"trying new location: moving from {origin} to {(c, r)}")
        return (c, r)

    def __random_location(self):
        assert self.data.model_width is not None
        assert self.data.model_height is not None
        origin = random_coords(
            Distribution.Uniform,
            self.data.model_width * self.data.model_height,
        )

        return np.unravel_index(origin, (self.data.model_height, self.data.model_width))  # type: ignore

    def spotlight_search(self, origin=None):
        if origin is None:
            centre = self.__random_location()
        else:
            centre = origin

        ret, resp, conf = self._Explanation__spatial(  # type: ignore
            centre=centre, expansion_limit=self.args.no_expansions
        )

        steps = 0
        while ret == SpatialSearch.NotFound and steps < self.args.max_spotlight_budget:
            if self.args.spotlight_objective_function == "none":
                centre = self.__random_location()
                ret, resp, conf = self._Explanation__spatial(  # type: ignore
                    centre=centre, expansion_limit=self.args.no_expansions
                )
            else:
                new_resp = 0.0
                while new_resp < resp:
                    centre = self.__random_step_from(
                        centre,
                        self.data.model_height,
                        self.data.model_width,
                        step=self.args.spotlight_step,
                    )
                    ret, new_resp, conf = self._Explanation__spatial(  # type: ignore
                        centre=centre, expansion_limit=self.args.no_expansions
                    )
                    if ret == SpatialSearch.Found:
                        return conf
                ret, resp, conf = self._Explanation__spatial(  # type: ignore
                    centre=centre, expansion_limit=self.args.no_expansions
                )
            steps += 1
        return conf



#!/usr/bin/env python
from typing import List, Optional
import numpy as np
import torch as tt
from PIL import Image  # type: ignore

try:
    from anytree.cachedsearch import find
except ImportError:
    from anytree.search import find

import matplotlib.pyplot as plt
from rex_xai.box import Box
from rex_xai.prediction import Prediction
from rex_xai.logger import logger
from rex_xai.input_data import Data
from rex_xai._utils import add_boundaries, set_boolean_mask_value

__combinations = [
    [
        0,
    ],
    [
        1,
    ],
    [
        2,
    ],
    [
        3,
    ],
    [0, 1],
    [0, 2],
    [0, 3],
    [1, 2],
    [1, 3],
    [2, 3],
    [0, 1, 2],
    [0, 1, 3],
    [0, 2, 3],
    [1, 2, 3],
]


def _apply_to_data(mask, data: Data, masking_func):
    if isinstance(masking_func, (float, int)):
        res = tt.where(mask, data.data, masking_func)  # type: ignore
        return res
    if callable(masking_func):
        return masking_func(mask, data.data)

    logger.warning("applying default masking value of 0")
    return tt.where(mask, data.data, 0)  # type: ignore


def get_combinations():
    return __combinations


class Mutant:
    def __init__(self, data: Data, static, active, masking_func) -> None:
        self.shape = tuple(
            data.model_shape[1:]
        )  # the first element of shape is the batch information, so we drop that
        self.mode = data.mode
        self.channels: int = (
            data.model_channels if data.model_channels is not None else 1
        )
        self.order = data.model_order
        self.mask = tt.zeros(self.shape, dtype=tt.bool, device=data.device)
        self.static = static
        self.active = active
        self.prediction: Optional[Prediction] = None
        self.passing = False
        self.masking_func = masking_func
        self.depth = 0

    def __repr__(self) -> str:
        return f"ACTIVE: {self.active}, PREDICTION: {self.prediction}, PASSING: {self.passing}"

    def get_name(self):
        return self.active

    def update_status(self, target):
        if self.prediction is not None:
            if target.classification == self.prediction.classification:
                self.passing = True

    def get_length(self):
        return len(self.active.split("_"))

    def get_active_boxes(self):
        return self.active.split("_")

    def area(self) -> int:
        """Return the total area *not* concealed by the mutant."""
        tensor = tt.count_nonzero(self.mask)
        if tensor.numel() == 0 or tensor is None:
            return 0
        else:
            return int(tensor.item()) // self.channels

    def set_static_mask_regions(self, names, search_tree):
        for name in names:
            box = find(search_tree, lambda node: node.name == name)
            if box is not None:
                self.depth = max(self.depth, box.depth)
                self.set_mask_region_to_true(box)

    def set_active_mask_regions(self, boxes: List[Box]):
        for box in boxes:
            self.depth = max(self.depth, box.depth)
            self.set_mask_region_to_true(box)

    def set_mask_region_to_true(self, box: Box):
        set_boolean_mask_value(self.mask, self.mode, self.order, box)

    def apply_to_data(self, data: Data):
        return _apply_to_data(self.mask, data, self.masking_func)

    def save_mutant(self, data: Data, name=None, segs=None):
        if data.mode in ("RGB", "L"):
            m = np.array(data.input.resize((data.model_height, data.model_width)))
            mask = self.mask.squeeze().detach().cpu().numpy()
            if data.transposed and data.mode == "RGB":
                # if transposed, we have C * H * W, so change that to H * W * C
                m = np.where(mask, m.transpose((2, 0, 1)), 0)
                m = m.transpose((1, 2, 0))
            else:
                # TODO m = m.transpose((0, 2, 1))
                m = np.where(mask, m, 255)
            # draw on the segment_mask, if available
            if segs is not None:
                m = add_boundaries(m, segs)
            img = Image.fromarray(m, data.mode)
            if name is not None:
                img.save(name)
            else:
                img.save(f"{self.get_name()}.png")
        # spectral or time series data
        if data.mode == "spectral":
            m = self.apply_to_data(data)
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.plot(m[0][0].detach().cpu().numpy())
            plt.savefig(f"{self.get_name()}.png")
        # 3d image
        if data.mode == "voxel":
            pass



#!/usr/bin/env python3

"""image generation functions"""

from PIL import Image, ImageDraw
import os

import torch as tt
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from scipy.ndimage import center_of_mass
from skimage.segmentation import slic
from torch import Tensor

from rex_xai.prediction import Prediction
from rex_xai.config import CausalArgs
from rex_xai.resp_maps import ResponsibilityMaps
from rex_xai.input_data import Data
from rex_xai._utils import add_boundaries
from rex_xai.logger import logger


def plot_curve(curve, chunk_size, style="insertion", destination=None):
    # TODO check that this code still works
    """plots an insertion/deletion curve of a responsibility map"""
    curve = np.array(curve)
    x = np.arange(0, len(curve))
    x *= chunk_size

    if style == "insertion":
        plt.plot(x, curve)
        plt.fill_between(x, curve, alpha=0.3)
        area = np.trapz(curve) / len(curve)
        plt.title(f"AUC for normalised insertion curve: {area:5.4f}")
        plt.xlabel("no. pixels")
        plt.ylabel("confidence")
    if style == "deletion":
        curve = 1.0 - curve
        plt.plot(x, curve)
        plt.fill_between(x, curve, alpha=0.3)
        area = np.trapz(curve) / len(curve)
        plt.title(f"AUC for normalised deletion curve: {area:5.4f}")
        plt.xlabel("no. pixels")
        plt.ylabel("confidence")

    if style == "both":
        fig, ax = plt.subplots(1, 2)
        fig.tight_layout()
        ax[0].plot(x, curve)
        ax[0].fill_between(x, curve, alpha=0.3)
        area = np.trapz(curve) / len(curve)
        ax[0].set_xlabel("no pixels")
        ax[0].set_ylabel("confidence")
        ax[0].set_title(f"AUC\ninsertion: {area}")

        curve = 1.0 - curve
        ax[1].plot(x, curve)
        ax[1].fill_between(x, curve, alpha=0.3)
        area = np.trapz(curve) / len(curve)
        ax[1].set_xlabel("no pixels")
        ax[1].set_ylabel("confidence")
        ax[1].set_title(f"AUC\ndeletion: {area}")

        fig.suptitle("Insertion/Deletion Curve")
        plt.subplots_adjust(top=0.85)

    if destination is None:
        plt.show()
    else:
        plt.savefig(destination, bbox_inches="tight", dpi=300, pad_inches=0)


def plot_3d(path, ranking, ogrid, norm=255.0):
    """plots a 3d grid in matplotlib given an image <path>
    If <path> is greyscale or RGBA, it is converted to RGB for plotting.
    """
    img = Image.open(path).convert("RGB")
    img = img.resize((ranking.shape[0], ranking.shape[1]))
    img = np.asarray(img)

    img = img / norm  # type: ignore
    if ogrid:
        x, y = np.ogrid[0 : img.shape[0], 0 : img.shape[1]]
    else:
        x, y = np.meshgrid(
            np.arange(0, ranking.shape[0], 1), np.arange(0, ranking.shape[1], 1)
        )
    return img, x, y


# code from https://stackoverflow.com/questions/42481203/heatmap-on-top-of-image
def _transparent_cmap(cmap, N=255):
    "Copy colormap and set alpha values"
    mycmap = cmap
    mycmap._init()
    mycmap._lut[:, -1] = np.linspace(0, 0.8, N + 4)
    return mycmap


def heatmap_plot(data: Data, resp_map, colour, path=None):
    if data.mode in ("RGB", "L"):
        mycmap = _transparent_cmap(mpl.colormaps[colour])
        background = data.input.resize(
            (data.model_height, data.model_width)
        )  # TODO check these dimensions
        y, x = np.mgrid[0 : data.model_height, 0 : data.model_width]
        fig, ax = plt.subplots(1, 1)
        ax.imshow(background)
        ax.contourf(x, y, resp_map, 15, cmap=mycmap)
        plt.axis("off")
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        if path is not None:
            plt.savefig(path, bbox_inches="tight", dpi=300, pad_inches=0)
        else:
            plt.show()


def contour_plot(path, maps: ResponsibilityMaps, target, levels=30, destination=None):
    """plots a contour plot"""
    pass


def __group_spectral_parts(explanation):
    # coords = tt.where(explanation)[0]

    coords = np.where(explanation.detach().cpu().numpy())[0]

    res = []
    local = [coords[0]]
    p = 1
    while p < len(coords):
        if coords[p] == coords[p - 1] + 1:
            local.append(coords[p])
        else:
            res.append(local)
            local = [coords[p]]
        p += 1
    res.append(local)

    return res


def spectral_plot(explanation, data: Data, ranking, colour, extra=True, path=None):
    if isinstance(ranking, tt.Tensor):
        ranking = ranking.detach().cpu().numpy()
    if isinstance(explanation, tt.Tensor):
        explanation = explanation.detach().cpu().numpy()

    explanation = explanation.squeeze()
    if extra:
        fig, axs = plt.subplots(nrows=2, ncols=1, sharex=True)
        ax = axs[0]
        axs[1].plot(ranking[0])
        axs[1].set_ylabel("Responsibility")
        axs[1].set_xlabel("Wavenumber")
    else:
        fig, ax = plt.subplots(nrows=1, ncols=1)

    raw_data = data.data[0, 0, :].detach().cpu().numpy()  # type: ignore
    d_min = np.min(raw_data)  # type: ignore
    d_max = np.max(raw_data)  # type: ignore

    # if the spectrum hasn't been base shifted to 0, then we do it to make plotting easier,
    # but we will lie about it on the y axis
    if d_min < 0:
        raw_data += np.abs(d_min)
        y_dmin = np.floor(d_min)
        y_dmax = np.ceil(d_max)
        ytx = np.abs(y_dmin) + y_dmax
        ticks = np.arange(0, ytx)
        labels = [str(x + y_dmin) for x in ticks]
        ax.set_yticks(ticks, labels=labels)
        ranking = np.repeat(ranking, len(labels), axis=0)

    ax.plot(raw_data)  # type: ignore
    ax.set_ylabel("Wave Intensity")

    k = data.target.classification  # type: ignore
    confidence = data.target.confidence  # type: ignore
    fig.suptitle(
        f"Spectrum and Responsibility\nTarget: {k}\nconfidence {confidence:5.4f}"
    )

    mycmap = _transparent_cmap(mpl.colormaps[colour])

    ranking = ranking / np.max(np.abs(ranking))

    c = ax.pcolormesh(ranking, cmap=mycmap)
    if not extra:
        # only plot the colorbar is we are not plotting a separate responsibility plot.
        fig.colorbar(c, ax=ax)

    fig.tight_layout()

    # coords = __group_spectral_parts(explanation)
    #
    # for rect in coords:
    #     rectangle = Rectangle(
    #         (rect[0], 0), rect[-1] - rect[0], 3, alpha=0.3, color="red"
    #     )
    #     axs[0].add_patch(rectangle)

    if path is None:
        plt.show()
    else:
        plt.savefig(path, dpi=300)


def surface_plot(
    args: CausalArgs,
    resp_map: np.ndarray,
    target: Prediction,
    path=None,
):
    """plots a 3d surface plot"""
    img, _x, _y = plot_3d(args.path, resp_map, True)
    fig = plt.figure()

    # TODO enable visualisation of sub-responsibility maps
    keys = [target.classification]

    rows, cols = 1, len(keys)

    # currently this for loop does nothing as keys always has len = 1
    # if passing in multiple responsibility maps, would need to use
    # ranking = resp_maps.get(k) for each iteration
    for i, k in enumerate(keys):
        ranking = resp_map
        if ranking is not None:
            ax = fig.add_subplot(rows, cols, i + 1, projection="3d")

            ax.plot_surface(  # type: ignore
                _x, _y, np.atleast_2d(0), rstride=5, cstride=5, facecolors=img
            )
            ax.plot_surface(  # type: ignore
                _x,
                _y,
                ranking,
                alpha=0.4,
                cmap=mpl.colormaps[args.heatmap_colours],
            )
            if args.info:
                # confidence = 0.0
                # if k == target.classification:
                confidence = target.confidence
                # else:
                #     for p in args.extra_targets:
                #         if p.pred == k:
                #             confidence = p.conf
                #             break

                try:
                    x, y = center_of_mass(ranking)
                    x = int(round(x))  # type: ignore
                    y = int(round(y))  # type: ignore
                    z = ranking[x, y]  # type: ignore
                    ax.scatter(x, y, z, color="b")

                    try:
                        lang = os.environ["LANG"]
                    except KeyError:
                        lang = "en_US.UTF-8"
                    if "GB" in lang:
                        ax.text(x, y, z, s="centre of mass")  # type: ignore
                    else:
                        ax.text(x, y, z, s="center of mass")  # type: ignore
                except ValueError:
                    pass

                loc = np.unravel_index(np.argmax(ranking), ranking.shape)
                ax.scatter(loc[0], loc[1], ranking[loc[0], loc[1]], color="r")
                ax.text(loc[0], loc[1], ranking[loc[0], loc[1]], s="max point")  # type: ignore
                if k == target:
                    plt.title(f"Target: {k}\nconfidence {confidence:5.4f}")
                else:
                    plt.title(f"Submap for {k}\nconfidence {confidence:5.4f}")

        if path is None:
            plt.show()
        else:
            plt.savefig(path, bbox_inches="tight", dpi=300)


def overlay_grid(img, step_count=10):
    draw = ImageDraw.Draw(img)

    y_start = 0
    y_end = img.height
    step_size = int(img.width / step_count)

    for x in range(0, img.width, step_size):
        line = ((x, y_start), (x, y_end))
        draw.line(line, fill=128)

    x_start = 0
    x_end = img.width

    for y in range(0, img.height, step_size):
        line = ((x_start, y), (x_end, y))
        draw.line(line, fill=128)

    del draw

    return img


def remove_background(data: Data, resp_map: np.ndarray) -> np.ndarray:
    """Remove the background from the responsibility map if set in the Data object"""
    if data.mode == "voxel":
        if len(data.data.shape) == 4:
            data_m = data.data[0, :, :, :]
        else:
            data_m = data.data
    else:
        data_m = data.data # need to check for other modes
    # Set background to minimum value in the responsibility map if set in the Data object
    if data.background is not None and data.background is int or float:
        background = np.where(
            data_m == data.background
        )  # Threshold for background voxels
        resp_map[background] = np.min(resp_map)
    elif data.background is not None and data.background is tuple:
        # Background is a range of values so (x , y) where x is the lower bound and y is the upper bound
        background = np.where(
            (data_m >= data.background[0]) & (data_m <= data.background[1])
        )
        resp_map[background] = np.min(resp_map)
    elif data.background is not None:
        logger.warn(
            "Background is not set correctly, please check the value. "
            "Background value must be an int, float or tuple defining the range of values for the background."
        )

    return resp_map


def voxel_plot(args: CausalArgs, resp_map: Tensor, data: Data, path=None):
    """
    Plot a 3D voxel plot of the responsibility map using plotly.
    - Assumes the data is greyscale
    Produces an interactive 3D plot of the data and the responsibility map.
    """
    try:
        import plotly.graph_objs as go
        from dash import Dash, dcc, html, Input, Output
    except ImportError as e:
        logger.error(f"Plotly failed to import caused by {e}.")
        return

    if len(data.data.shape) == 4:
        data_m = data.data[0, :, :, :]
    else:
        data_m = data.data
    if isinstance(resp_map, tt.Tensor):
        resp_map = resp_map.squeeze().detach().cpu().numpy()

    if isinstance(data_m, tt.Tensor):
        data_m = data_m.squeeze().detach().cpu().numpy()
        tt.tensor(data_m, dtype=tt.float32)

    maps: np.ndarray = resp_map
    resp_map = remove_background(data, maps)

    # Normalize the data
    data_m = (data_m - np.min(data_m)) / (np.max(data_m) - np.min(data_m))
    resp_map = (resp_map - np.min(resp_map)) / (np.max(resp_map) - np.min(resp_map))

    # Check if both data and responsibility map have the same range of values
    assert np.min(data_m) == np.min(resp_map) and np.max(data_m) == np.max(resp_map), (
        "Data and Responsibility map must have the same range of values!"
    )

    assert data_m.shape == maps.shape, (
        "Data and Responsibility map must have the same shape!"
    )

    x_max, y_max, z_max = data_m.shape

    colourscales = list(mpl.colormaps.keys())

    app = Dash(__name__)
    x_slice = go.Figure()
    y_slice = go.Figure()
    z_slice = go.Figure()

    app.layout = html.Div([
        html.Div([
            # X Slice
            html.Div([
                html.Label("X Slice", style={"font-weight": "bold", "margin-bottom": "10px"}),
                dcc.Graph(id="x-slice", style={"width": "100%", "height": "auto", "max-width": "400px"}),
                dcc.Slider(0, x_max - 1, 1, value=x_max // 2, id="x-slider",
                           marks={0: "0", x_max - 1: str(x_max - 1)},
                           vertical=True, tooltip={"always_visible": True}),
            ], style={"display": "flex", "align-items": "center", "gap": "20px", "flex": "1"}),

            # Y Slice
            html.Div([
                html.Label("Y Slice", style={"font-weight": "bold", "margin-bottom": "10px"}),
                dcc.Graph(id="y-slice", style={"width": "100%", "height": "auto", "max-width": "400px"}),
                dcc.Slider(0, y_max - 1, 1, value=y_max // 2, id="y-slider",
                           marks={0: "0", y_max - 1: str(y_max - 1)},
                           vertical=True, tooltip={"always_visible": True}),
            ], style={"display": "flex", "align-items": "center", "gap": "20px", "flex": "1"}),

        ], style={"display": "flex", "justify-content": "center", "gap": "40px"}),

        # Z Slice
        html.Div([
            html.Label("Z Slice", style={"font-weight": "bold", "margin-bottom": "10px"}),
            dcc.Graph(id="z-slice", style={"width": "100%", "height": "auto", "max-width": "400px"}),
            dcc.Slider(0, z_max - 1, 1, value=z_max // 2, id="z-slider",
                       marks={0: "0", z_max - 1: str(z_max - 1)},
                       vertical=True, tooltip={"always_visible": True}),
        ], style={"display": "flex", "align-items": "center", "gap": "20px", "margin-top": "40px"}),

        # Opacity Slider
        html.Div([
            html.Label("Opacity"),
            dcc.Slider(0, 1, 0.1, value=0.5, id="opacity-slider",
                       tooltip={"always_visible": True}, marks={0: "0", 1: "1"},
                       vertical=True),
            html.Label("Heatmap Colours"),
            dcc.Dropdown(id="heatmap-colours", options=colourscales, value=args.heatmap_colours)
        ], style={"position": "absolute", "top": "10%", "right": "10%", "width": "100px", "outline": "1px solid grey"}),

    ], style={"width": "3000x", "height": "20px", "margin": "auto", "padding": "20px", "display": "flex",
              "flex-direction": "column", "gap": "40px", "outline": "1px solid grey"})

    @app.callback(
        Output("x-slice", "figure"),
        Output("y-slice", "figure"),
        Output("z-slice", "figure"),
        Input("x-slider", "value"),
        Input("y-slider", "value"),
        Input("z-slider", "value"),
        Input("opacity-slider", "value"),
        Input("heatmap-colours", "value")
    )
    def update_slices(x_idx, y_idx, z_idx, opacity, heatmap_colours=args.heatmap_colours):
        # X-Slice (YZ plane)
        x_slice.add_trace(
            go.Heatmap(z=data_m[x_idx, :, :], colorscale="gray_r", name="Data", zmin=0, zmax=1, showscale=False))
        x_slice.add_trace(
            go.Heatmap(z=resp_map[x_idx, :, :], colorscale=heatmap_colours, opacity=opacity, name="Resp Map", zmin=0, zmax=1))
        x_slice.update_layout(title=f"YZ Plane at {x_idx}")

        # Y-Slice (XZ plane)
        y_slice.add_trace(
            go.Heatmap(z=data_m[:, y_idx, :], colorscale="gray_r", name="Data", zmin=0, zmax=1, showscale=False))
        y_slice.add_trace(
            go.Heatmap(z=resp_map[:, y_idx, :], colorscale=heatmap_colours, opacity=opacity, name="Resp Map", zmin=0, zmax=1))
        y_slice.update_layout(title=f"XZ Plane at {y_idx}")

        # Z-Slice (XY plane)
        z_slice.add_trace(
            go.Heatmap(z=data_m[:, :, z_idx], colorscale="gray_r", name="Data", zmin=0, zmax=1, showscale=False))
        z_slice.add_trace(
            go.Heatmap(z=resp_map[:, :, z_idx], colorscale=heatmap_colours, opacity=opacity, name="Resp Map", zmin=0, zmax=1))
        z_slice.update_layout(title=f"XY Plane at {z_idx}")

        return x_slice, y_slice, z_slice

    if path is None:
        app.run_server(debug=False, use_reloader=False)
    else:
        # Create a sub figure for each slice and save it for now
        path = path.split(".")[0]
        x_slice.write_image(f"{path}_x_slice.png")
        y_slice.write_image(f"{path}_y_slice.png")
        z_slice.write_image(f"{path}_z_slice.png")



def __transpose_mask(explanation, mode, transposed):
    mask = None
    if transposed:
        if mode == "RGB":
            mask = explanation.squeeze().detach().cpu().numpy().transpose((1, 2, 0))
        elif mode == "L":
            mask = explanation.squeeze(0).detach().cpu().numpy().transpose((2, 1, 0))
            mask = np.repeat(mask, 3, axis=-1)
    else:
        mask = explanation.squeeze(0).detach().cpu().numpy()

    return mask


def generate_colours(n, colourmap):
    """
    Generate n evenly spaced RGB colours from a matplotlib colourmap.
    """
    space = np.linspace(0, 1, n)
    colour_space = mpl.colormaps[colourmap].resampled(n)
    # colour space returns colours in RGBA space, so we drop the A at the end
    rgb_colours = [colour_space(i)[:-1] for i in space]
    return rgb_colours


def make_composite_mask(explanations):
    """
    Creates a composite mask from a list of masks.
    """
    composite_mask = None
    for explanation in explanations:
        if composite_mask is None:
            composite_mask = explanation
        else:
            composite_mask = np.where(explanation, 1, composite_mask)

    return composite_mask


def apply_boundaries_to_image(image, explanations, colours):
    """
    Draws the boundaries of the explanations on the image, using the provided colours.
    """
    for i in range(len(explanations)):
        explanation = explanations[i]

        if explanation.shape[0] == 3:
            explanation = explanation[0, :, :]
        else:
            explanation = explanation[:, :, 0]  # type: ignore

        image = add_boundaries(image, explanation, colour=colours[i])

    return image


def get_img_as_array(data):
    """
    Return original input image as a numpy array, resized to match model size and converted to RGB if necessary.
    """
    if data.mode == "RGB" or data.mode == "L":
        if data.mode == "L":
            img = data.input.convert("RGB").resize(
                (data.model_height, data.model_width)
            )
        else:
            img = data.input.resize((data.model_height, data.model_width))
    else:
        raise NotImplementedError

    return np.array(img)


def save_multi_explanation(
    explanations, data, args: CausalArgs, clause=None, path=None
):
    if data.mode == "RGB" or data.mode == "L":
        img = get_img_as_array(data)
    else:
        logger.warning("we do not yet handle multiple explanations for non-images")
        raise NotImplementedError

    if img is not None:
        rgb_colours = generate_colours(args.spotlights, args.heatmap_colours)

        if clause is not None:
            if isinstance(clause, int):
                explanations_subset = [explanations[clause]]
                colours_subset = [rgb_colours[clause]]
            else:
                explanations_subset = [explanations[c] for c in clause]
                colours_subset = [rgb_colours[c] for c in clause]
            explanations_subset = [
                __transpose_mask(explanation, data.mode, data.transposed)
                for explanation in explanations_subset
            ]
            composite_mask = make_composite_mask(explanations_subset)

            img = apply_boundaries_to_image(img, explanations_subset, colours_subset)

            if composite_mask is not None:
                cover = np.where(composite_mask, img, args.colour)
                cover = Image.fromarray(cover, data.mode)
                img = Image.fromarray(img, data.mode)
                out = Image.blend(cover, img, args.alpha)

                if path is None:
                    return out
                else:
                    out.save(path)


def save_image(explanation, data: Data, args: CausalArgs, path=None):
    mask = None
    if data.mode == "RGB" or data.mode == "L":
        if data.mode == "L":
            img = data.input.convert("RGB").resize(
                (data.model_height, data.model_width)
            )
        else:
            img = data.input.resize((data.model_height, data.model_width))

        mask = __transpose_mask(explanation, data.mode, data.transposed)

        if mask is not None:
            if args.raw:
                out = np.where(mask, img, 0).squeeze(
                    0
                )  # 0 used to mask image with black
                out = Image.fromarray(out, data.mode)

            else:
                exp = np.where(mask, img, args.colour)
                exp = Image.fromarray(exp, "RGB")
                out = Image.blend(exp, img, args.alpha)

                if args.mark_segments:
                    segs = slic(np.array(img))
                    m = add_boundaries(np.array(img), segs)
                    marked = Image.fromarray(m, data.mode)
                    out = Image.blend(out, marked, args.alpha)

                if args.grid:
                    out = overlay_grid(out)

                if args.resize:
                    out = out.resize(data.input.size)

            if path is not None:
                out.save(path)

            return out
    elif data.mode == "voxel":
        data_m: np.ndarray = data.data
        if isinstance(explanation, tt.Tensor):
            explanation = explanation.squeeze().detach().cpu().numpy()
        else:
            explanation = explanation.squeeze()
        data_m = data_m[0, :, :, :]  # Remove batch dimension

        explanation = remove_background(data, explanation)

        num_slices = 10
        fig, axes = plt.subplots(3, num_slices, figsize=(15, 6))

        for axis, index in enumerate(explanation.shape):
            slice_indices = np.linspace(0, axis - 1, num_slices, dtype=int)
            for i, slice_index in enumerate(slice_indices):
                ax = axes[axis, i]
                data_slice = np.take(data_m, slice_index, axis=axis)
                resp_slice = np.take(explanation, slice_index, axis=axis)
                ax.imshow(data_slice, cmap="gray")
                ax.imshow(resp_slice, cmap=args.heatmap_colours, alpha=0.4)
                ax.axis("off")

        plt.tight_layout()

        if args.output is not None:
            plt.savefig(args.output)
        else:
            plt.show()


def plot_image_grid(images, ncols=None):
    # adapted from: https://stackoverflow.com/questions/41793931/plotting-images-side-by-side-using-matplotlib/66961099#66961099
    """Plot a grid of images"""
    if not ncols:
        factors = [i for i in range(1, len(images) + 1) if len(images) % i == 0]
        ncols = factors[len(factors) // 2] if len(factors) else len(images) // 4 + 1
    nrows = int(len(images) / ncols) + int(len(images) % ncols)
    imgs = [images[i] if len(images) > i else None for i in range(nrows * ncols)]
    f, axes = plt.subplots(nrows, ncols, figsize=(3 * ncols, 2 * nrows))
    axes = axes.flatten()[: len(imgs)]
    for img, ax in zip(imgs, axes.flatten()):
        ax.imshow(img)
        ax.set_axis_off()
    f.tight_layout()



#!/usr/bin/env python
import torch as tt
import numpy as np
from scipy.ndimage import gaussian_filter


def __split_groups(neg_mask):
    # for some reason, it's much faster to do this on the cpu with numpy
    # than it is to use tensor_split
    return np.split(neg_mask, np.where(np.diff(neg_mask) > 1)[0] + 1)


def spectral_occlusion(mask: tt.Tensor, data: tt.Tensor, noise=0.02, device="cpu"):
    """Linear interpolated occlusion for spectral data, with optional added noise.

    @param mask: boolean valued NDArray
    @param data: data to be occluded
    @param noise: parameter for optional gaussian noise.
        Set to 0.0 if you want simple linear interpolation

    @return torch.Tensor
    """
    neg_mask = tt.where(mask == 0)[0]
    split = __split_groups(neg_mask.detach().cpu().numpy())

    # strangely, this all seems to run faster if we do it on the cpu.
    # TODO Needs further investigation
    local_data = np.copy(data.detach().cpu().numpy())

    for s in split:
        if len(s) <= 1:
            return tt.from_numpy(local_data).to(device)
        start = s[0]
        stop = s[-1]
        dstart = data[:, :, start][0][0].item()
        dstop = data[:, :, stop][0][0].item()
        interp = tt.linspace(dstart, dstop, stop - start)
        if noise > 0.0:
            interp += np.random.normal(0, noise, len(interp))

        local_data[0, 0, start:stop] = interp

    return tt.from_numpy(local_data).to(device)


# Occlusions such as beach, sky, and other context-based occlusions


# Medical-based occlusions could be a CT scan of a healthy patient
def context_occlusion(mask: tt.Tensor, data: tt.Tensor, context: tt.Tensor, noise=0.5):
    """Context based occlusion with optional added noise.

    @param mask: boolean valued NDArray
    @param data: data to be occluded
    @param context: data to be used as occlusion e.g. CT scan of a healthy patient or a road
    @param noise: parameter for optional gaussian noise.
        Set to 0.0 for no noise

    @return torch.Tensor
    """
    if noise > 0.0:
        context = tt.tensor(gaussian_filter(context, sigma=noise), dtype=tt.float32)
    return tt.where(mask == 0, context, data)



"""distributions module"""

from typing import Optional, Tuple
from enum import Enum
import numpy as np
from scipy.stats import binom, betabinom
from rex_xai.logger import logger

Distribution = Enum("Distribution", ["Binomial", "Uniform", "BetaBinomial", "Adaptive"])


def _betabinom2d(height, width, alpha, beta):
    bx = betabinom(width, alpha, beta)
    by = betabinom(height, alpha, beta)

    w = np.array([bx.pmf(i) for i in range(0, width)])  # type: ignore
    h = np.array([by.pmf(i) for i in range(0, height)])  # type: ignore

    w = np.expand_dims(w, axis=0)
    h = np.expand_dims(h, axis=0)

    u = (h.T * w / np.sum(h.T * w)).ravel()
    p = np.random.choice(np.arange(0, len(u)), p=u)
    return p


def _blend(dist, alpha, base):
    pmf = np.array([base.pmf(x) for x in range(0, len(dist))])
    blend = ((1.0 - alpha) * pmf) + (alpha * dist)
    blend /= np.sum(blend)
    return blend


def _2d_adaptive(map, args: Tuple[int, int, int, int], alpha=0.0, base=None) -> int:
    # if the map exists and is not 0.0 everywhere...
    if map is not None and np.max(map) > 0.0:
        s = map[args[0] : args[1], args[2] : args[3]]
        sf = np.ndarray.flatten(s)
        # sf = np.max(sf) - sf
        sf /= np.sum(sf)

        # base = betabinom(0, len(sf), 1.1, 1.1)
        # if base is not None:
        #     sf = _blend(alpha, base)
        pos = np.random.choice(np.arange(0, len(sf)), p=sf)
        return pos

    # if the map is empty or doesn't exist, return uniform
    return np.random.randint(1, (args[1] - args[0]) * (args[3] - args[2]))


def str2distribution(d: str) -> Distribution:
    """converts string into Distribution enum"""
    if d == "binom":
        return Distribution.Binomial
    elif d == "uniform":
        return Distribution.Uniform
    elif d == "betabinom":
        return Distribution.BetaBinomial
    elif d == "adaptive":
        return Distribution.Adaptive
    else:
        logger.warning(
            "Invalid distribution '%s', reverting to default value Distribution.Uniform",
            d,
        )
    return Distribution.Uniform


def random_coords(d: Optional[Distribution], *args, map=None) -> Optional[int]:
    """generates random coordinates given a distribution and args"""

    try:
        if d == Distribution.Adaptive:
            return _2d_adaptive(map, args[0])

        if d == Distribution.Uniform:
            return np.random.randint(1, args[0])  # type: ignore

        if d == Distribution.Binomial:
            start, stop, *dist_args = args[0]
            return binom(stop - start - 1, dist_args).rvs() + start

        if d == Distribution.BetaBinomial:
            return _betabinom2d(args[1], args[2], args[3][0], args[3][1])

        return None
    except ValueError:
        return None



#!/usr/bin/env python
import numpy as np
from typing import List

from typing import Optional

try:
    from anytree.cachedsearch import find
except ImportError:
    from anytree import find

from rex_xai.box import Box
from rex_xai.config import CausalArgs
from rex_xai.mutant import Mutant
from rex_xai.input_data import Data
from rex_xai.logger import logger
from rex_xai._utils import ReXMapError


class ResponsibilityMaps:
    def __init__(self) -> None:
        self.maps = {}
        self.counts = {}

    def __repr__(self) -> str:
        return str(self.counts)

    def get(self, k, increment=False):
        try:
            if increment:
                self.counts[k] += 1  # type: ignore
            return self.maps[k]
        except KeyError:
            return

    def new_map(self, k: int, height, width, depth=None):
        if depth is not None:
            self.maps[k] = np.zeros((height, width, depth), dtype="float32")
            self.counts[k] = 1
        else:
            self.maps[k] = np.zeros((height, width), dtype="float32")
            self.counts[k] = 1

    def items(self):
        return self.maps.items()

    def keys(self):
        return self.maps.keys()

    def len(self):
        return len(self.maps)

    def merge(self, maps):
        for k, v in maps.items():
            if k in self.maps:
                self.maps[k] += v
            else:
                self.maps[k] = v

    def responsibility(self, mutant: Mutant, args: CausalArgs):
        responsibility = np.zeros(4, dtype=np.float32)
        parts = mutant.get_active_boxes()
        r = 1 / len(parts)
        for p in parts:
            i = np.uint(p[-1])
            if (
                args.weighted
                and mutant.prediction is not None
                and mutant.prediction.confidence is not None
            ):
                responsibility[i] += r * mutant.prediction.confidence
            else:
                responsibility[i] += r
        return responsibility

    def update_maps(
        self, mutants: List[Mutant], args: CausalArgs, data: Data, search_tree
    ):
        """Update the different responsibility maps with all passing mutants <mutants>
        @params mutants: list of mutants
        @params args: causal args
        @params data: data
        @params search_tree: tree of boxes

        Mutates in place, does not return a value
        """

        for mutant in mutants:
            r = self.responsibility(mutant, args)

            k = None
            # check that there is a prediction value
            if mutant.prediction is not None:
                k = mutant.prediction.classification
            # if there's no prediction value, raise an exception
            if k is None:
                raise ReXMapError("the provided mutant has no known classification")
            # check if k has been seen before and has a map. If k is new, make a new map
            if k not in self.maps:
                self.new_map(
                    k, data.model_height, data.model_width, data.model_depth
                )

            # get the responsibility map for k
            resp_map = self.get(k, increment=True)
            if resp_map is None:
                raise ValueError(
                    f"unable to open or generate a responsibility map for classification {k}"
                )

            # we only increment responsibility for active boxes, not static boxes
            for box_name in mutant.get_active_boxes():
                box: Optional[Box] = find(search_tree, lambda n: n.name == box_name)
                if box is not None and box.area() > 0:
                    index = np.uint(box_name[-1])
                    local_r = r[index]
                    if args.concentrate:
                        local_r *= 1.0 / box.area()
                        # Don't delete this code just yet as this is an alternative (less brutal)
                        # scaling strategy that needs further investigation
                        # scale = depth - 1
                        # local_r = 2**(local_r * scale)

                    if data.mode == "spectral":
                        section = resp_map[0, box.col_start : box.col_stop]
                    elif data.mode == "RGB":
                        section = resp_map[
                            box.row_start : box.row_stop,
                            box.col_start : box.col_stop,
                        ]
                    elif data.mode == "voxel":
                        section = resp_map[
                            box.row_start : box.row_stop,
                            box.col_start : box.col_stop,
                            box.depth_start : box.depth_stop,
                        ]
                    else:
                        logger.warning("not yet implemented")
                        raise NotImplementedError

                    section += local_r
            self.maps[k] = resp_map

    def subset(self, id):
        m = self.maps.get(id)
        c = self.counts.get(id)
        self.maps = {id: m}
        self.counts = {id: c}



#!/usr/bin/env python
import re
from typing import Optional

import torch as tt

from rex_xai import visualisation
from rex_xai._utils import SpatialSearch, get_map_locations, set_boolean_mask_value
from rex_xai.config import CausalArgs, Strategy
from rex_xai.input_data import Data
from rex_xai.logger import logger
from rex_xai.mutant import _apply_to_data
from rex_xai.resp_maps import ResponsibilityMaps


class Explanation:
    def __init__(
        self,
        maps: ResponsibilityMaps,
        prediction_func,
        data: Data,
        args: CausalArgs,
        run_stats: dict,
        keep_all_maps=False,
    ) -> None:
        if data.target is None or data.target.classification is None:
            raise (
                ValueError(
                    "Data must have `target` defined to create an Explanation object!"
                )
            )

        if keep_all_maps:
            self.maps = maps
        else:
            maps.subset(data.target.classification)
            self.maps = maps

        self.target_map = tt.from_numpy(maps.get(data.target.classification)).to(
            data.device
        )
        if self.target_map is None:
            raise ValueError(
                f"No responsibility map found for target {data.target.classification}!"
            )

        self.explanation: Optional[tt.Tensor] = None
        self.final_mask = None
        self.explanation_confidence = 0.0
        self.prediction_func = prediction_func
        self.data = data
        self.args = args
        self.run_stats = run_stats

    def __repr__(self) -> str:
        pred_func = repr(self.prediction_func)
        match_func_name = re.search(r"(<function .+) at", pred_func)
        if match_func_name:
            pred_func = match_func_name.group(1) + " >"

        run_stats = {k: round(v, 5) for k, v in self.run_stats.items()}

        exp_text = (
            "Explanation:"
            + f"\n\tCausalArgs: {type(self.args)}"
            + f"\n\tData: {self.data}"
            + f"\n\tprediction function: {pred_func}"
            + f"\n\tResponsibilityMaps: {self.maps}"
            + f"\n\trun statistics: {run_stats} (5 dp)"
        )

        if self.explanation is None or self.final_mask is None:
            return (
                exp_text
                + f"\n\texplanation: {self.explanation}"
                + f"\n\tfinal mask: {self.final_mask}"
                + f"\n\texplanation confidence: {self.explanation_confidence}"
            )
        else:
            return (
                exp_text
                + f"\n\texplanation: {type(self.explanation)} of shape {self.explanation.shape}"
                + f"\n\tfinal mask: {type(self.final_mask)} of shape {self.final_mask.shape}"
                + f"\n\texplanation confidence: {self.explanation_confidence:.5f} (5 dp)"
            )

    def extract(self, method: Strategy):
        self.blank()
        if method == Strategy.Global:
            self.__global()
        if method == Strategy.Spatial:
            if self.data.mode == "spectral":
                logger.warning(
                    "spatial search not yet implemented for spectral data, so defaulting to global search"
                )
                self.__global()
            else:
                _ = self.__spatial()

        if isinstance(self.final_mask, tt.Tensor):
            self.final_mask = self.final_mask.detach().cpu().numpy()
        if isinstance(self.target_map, tt.Tensor):
            self.target_map = self.target_map.detach().cpu().numpy()

    def blank(self):
        assert self.data.data is not None
        self.explanation = tt.zeros(
            self.data.data.shape, dtype=tt.bool, device=self.data.device
        )
        self.final_mask = None

    def set_to_true(self, coords, mask=None):
        if mask is not None:
            mask = set_boolean_mask_value(
                mask, self.data.mode, self.data.model_order, coords
            )

    def __global(self, map=None, wipe=False):
        if map is None:
            map = self.target_map
        ranking = get_map_locations(map)

        mutant = tt.zeros(
            self.data.model_shape[1:], dtype=tt.bool, device=self.data.device
        )
        masks = []

        limit = 0
        for i in range(0, len(ranking), self.args.chunk_size):
            chunk = ranking[i : i + self.args.chunk_size]
            limit += self.args.chunk_size
            for _, loc in chunk:
                self.set_to_true(loc, mutant)
            d = _apply_to_data(mutant, self.data, self.data.mask_value).squeeze(0)
            masks.append(d)
            if len(masks) == self.args.batch_size:
                preds = self.prediction_func(tt.stack(masks).to(self.data.device))
                for j, p in enumerate(preds):
                    if (
                        p.classification == self.data.target.classification
                        and p.confidence
                        >= self.data.target.confidence
                        * self.args.minimum_confidence_threshold
                    ):  #  type: ignore
                        logger.info(
                            "found an explanation of %d with %f confidence",
                            p.classification,
                            p.confidence,
                        )
                        self.explanation = masks[j]
                        self.explanation_confidence = p.confidence
                        self.final_mask = mutant.zero_()
                        for _, loc in ranking[:limit]:
                            self.set_to_true(loc, self.final_mask)
                        return p.confidence
                masks = []

    def __generate_circle_coordinates(self, centre, radius: int):
        assert self.data.model_height is not None
        assert self.data.model_width is not None
        Y, X = tt.meshgrid(
            tt.arange(0, self.data.model_height),
            tt.arange(0, self.data.model_width),
            indexing="ij",
        )

        dist_from_centre = tt.sqrt(
            (Y.to(self.data.device) - centre[0]) ** 2
            + (X.to(self.data.device) - centre[1]) ** 2
        )

        # this produces a H * W mask which can be using in conjunction with tt.where()
        circle_mask = dist_from_centre <= radius

        return circle_mask

    def __draw_circle(self, centre, start_radius=None):
        if start_radius is None:
            start_radius = self.args.spatial_initial_radius
        mask = tt.zeros(
            self.data.model_shape[1:], dtype=tt.bool, device=self.data.device
        )
        circle_mask = self.__generate_circle_coordinates(centre, start_radius)
        if self.data.model_order == "first":
            mask[:, circle_mask] = True
        else:
            mask[circle_mask, :] = True
        return start_radius, circle_mask, mask

    def compute_masked_responsibility(self, mask):
        try:
            masked_responsibility = tt.where(
                mask, self.target_map, self.data.mask_value
            )  # type: ignore
        except RuntimeError:
            masked_responsibility = tt.where(
                mask.permute((2, 0, 1)), self.target_map, self.data.mask_value
            )  # type: ignore
        except Exception as e:
            logger.fatal(e)
            exit()

        logger.debug("using %s", self.args.spotlight_objective_function)
        if self.args.spotlight_objective_function == "mean":
            return tt.mean(masked_responsibility).item()
        if self.args.spotlight_objective_function == "max":
            return tt.max(masked_responsibility).item()

        logger.warning(
            "unable to understand %s, so using mean for search",
            self.args.spotlight_objective_function,
        )
        return tt.mean(masked_responsibility).item()

    def __spatial(self, centre=None, expansion_limit=None):
        # we don't have a search location to start from, so we try to isolate one
        map = self.target_map
        if centre is None:
            centre = tt.unravel_index(tt.argmax(map), map.shape)  # type: ignore

        start_radius, circle, mask = self.__draw_circle(centre)

        if self.args.spotlight_objective_function == "none":
            masked_responsibility = None
        else:
            masked_responsibility = self.compute_masked_responsibility(mask)

        expansions = 0
        cutoff = (
            self.data.model_width * self.data.model_height * self.data.model_channels  # type: ignore
        )
        while tt.count_nonzero(mask) < cutoff:
            if expansion_limit is not None:
                if expansions >= expansion_limit and expansion_limit > 1:
                    logger.debug(
                        f"no explanation found after {expansion_limit} expansions"
                    )
                    return SpatialSearch.NotFound, masked_responsibility, None
            d = _apply_to_data(mask, self.data, self.data.mask_value)
            p = self.prediction_func(d)[0]
            if (
                p.classification == self.data.target.classification  # type: ignore
                and p.confidence
                >= self.data.target.confidence * self.args.minimum_confidence_threshold  # type: ignore
            ):
                conf = self.__global(map=tt.where(circle, map, 0))  # type: ignore
                return SpatialSearch.Found, masked_responsibility, conf
            start_radius = int(start_radius * (1 + self.args.spatial_radius_eta))
            _, circle, _ = self.__draw_circle(centre, start_radius)
            if self.data.model_order == "first":
                mask[:, circle] = True
            else:
                mask[circle, :] = True
            expansions += 1

    def save(self, path, mask=None, multi=None, multi_style="", clauses=None):
        # NOTE: the parameter multi_style="" is here simply to make overriding
        # the save function in MultiExplanation typecheck, same holds for clauses
        if self.data.mode in ("RGB", "L", "voxel"):
            if path is None:
                path = f"{self.data.target.classification}.png"  # type: ignore
            if mask is None:
                visualisation.save_image(
                    self.explanation, self.data, self.args, path=path
                )
            else:
                visualisation.save_image(mask, self.data, self.args, path=path)
        if self.data.mode == "spectral":
            visualisation.spectral_plot(
                self.explanation,
                self.data,
                self.target_map,
                self.args.heatmap_colours,
                path=path,
            )
        if self.data.mode == "tabular":
            pass

    def heatmap_plot(self, path=None):
        if self.data.mode in ("RGB", "L"):
            visualisation.heatmap_plot(
                self.data,
                self.target_map,
                self.args.heatmap_colours,
                path=path,
            )
        elif self.data.mode == "voxel":
            visualisation.voxel_plot(
                self.args,
                self.target_map,
                self.data,
                path=path,
            )
        else:
            return NotImplementedError

    def surface_plot(self, path=None):
        if self.data.mode in ("RGB", "L"):
            visualisation.surface_plot(
                self.args,
                self.target_map,  # type: ignore
                self.data.target,  #  type: ignore
                path=path,
            )
        elif self.data.mode == "voxel":
            logger.warning("Surface plot not available for voxel data using voxel plot instead")
            visualisation.voxel_plot(
                self.args,
                self.target_map,
                self.data,
                path=path,
            )
        else:
            return NotImplementedError

    def show(self, path=None):
        if self.data.mode in ("RGB", "L", "voxel"):
            out = visualisation.save_image(
                self.explanation, self.data, self.args, path=path
            )
            return out
        else:
            return NotImplementedError



#!/usr/bin/env python3
# from __future__ import annotations
"""main logical entrypoint for ReX."""

import copy
import os
import sys
import time
from typing import Tuple, List, Union

from scipy.io import loadmat
import numpy as np
import torch as tt
from PIL import Image
from sqlalchemy.orm import Session
from tqdm import trange  # type: ignore

from rex_xai.config import CausalArgs
from rex_xai.database import update_database
from rex_xai.evaluation import Evaluation
from rex_xai.multi_explanation import MultiExplanation
from rex_xai.extraction import Explanation
from rex_xai.input_data import Data
from rex_xai.logger import logger
from rex_xai.onnx import get_prediction_function
from rex_xai.resp_maps import ResponsibilityMaps
from rex_xai.responsibility import causal_explanation
from rex_xai.prediction import Prediction
from rex_xai._utils import Strategy, ReXScriptError


def try_preprocess(args: CausalArgs, model_shape: Tuple[int], device: tt.device):
    """Makes an attempt to preprocess input data as required for the model.

    Data preprocessing is based on file extension and (possibly) user-defined mode.
    File extensions in ``[".jpg", ".jpeg", ".png", ".tif", ".tiff"]`` are treated
    as images, ".npy" and ".mat" are treated as Numpy arrays, and ".nii" are treated as nifti files.
    For any other file extension, we create a ``Data`` object without pre-processing.

    Args:
        args: configuration values for ReX
        model_shape: shape of the input tensor of the model, as returned by
            :py:func:`~rex_xai.explanation.get_prediction_func_from_args()`
        device: as returned by :py:func:`~rex_xai._utils.get_device()`

    Returns:
      Data: the processed input data
    """
    _, ext = os.path.splitext(args.path)
    # args.path is an image
    if ext.lower() in [".jpg", ".jpeg", ".png", ".tif", ".tiff"]:
        # a simple sanity check: spectral_occlusion is not suitable for images
        # so remind the user to update rex.toml and set mask_value to 0 in the meantime
        try:
            if args.mask_value in ("tabular", "spectral"):  # type: ignore
                logger.warning(
                    f"{args.mask_value} is not suitable for images. Changing mask_value to 0"
                )
                args.mask_value = 0
        except AttributeError:
            pass

        img = Image.open(args.path)
        if img.mode == "RGBA":
            logger.warning("RGBA input image provided, converting to RGB")
            img = img.convert("RGB")

        data = Data(img, model_shape, device)
        data.generic_image_preprocess(means=args.means, stds=args.stds, norm=args.norm)

    # a numpy "npy" array or matlab "mat" file
    elif ext in (".npy", ".mat"):
        if args.mode in ("tabular", "spectral"):
            if ext == ".mat":
                raw_data = np.load(loadmat(args.path)["val"])
            else:
                raw_data = np.load(args.path)
            data = Data(raw_data, model_shape, mode=args.mode, device=device)
            data.data = data.generic_tab_preprocess()
        else:
            logger.fatal("we do not generically handle this datatype")
            return NotImplemented
    # nifti files for 3d data
    elif ext in ".nii":
        logger.fatal("we do not (yet) handle nifti files generically")
        return NotImplemented
    else:
        # we don't know what to do!
        data = Data(args.path, model_shape, mode=args.mode, device=device)

    return data


def load_and_preprocess_data(
    model_shape: Tuple[int], device: tt.device, args: CausalArgs
):
    """Loads input data from filepath and does preprocessing.

    Uses a custom preprocesssing function if this is defined in ``args.script.preprocess``,
    otherwise :py:func:`~rex_xai.explanation.try_preprocess()`.

    Args:
        model_shape: shape of the input tensor of the model, as returned by
            :py:func:`~rex_xai.explanation.get_prediction_func_from_args()`
        device: as returned by :py:func:`~rex_xai._utils.get_device()`
        args: configuration values for ReX

    Returns:
        Data: the processed input data

    """
    if args.script is not None:
        if hasattr(args.script, "preprocess"):
            data = args.script.preprocess(
                args.path, model_shape, device, mode=args.mode
            )
        else:
            raise ReXScriptError(
                f"{args.script_location} is missing a preprocess() function"
            )
    else:
        # no custom preprocessing, so we make our best guess as to what to do
        data = try_preprocess(args, model_shape, device)

    return data


def predict_target(data: Data, prediction_func) -> Prediction:
    """Predicts classification of input data, using given prediction function.

    Uses ``prediction_func`` to identify the classification of the input data and return
    this as the target classification for ReX.

    Args:
        data: processed input data object
        args: configuration values for ReX
        prediction_func: prediction function for the model

    Returns:
        Prediction: the predicted target classification and confidence
    """
    target = prediction_func(data.data, None)

    if isinstance(target, list):
        target = target[0]

    if target is not None:
        logger.info(
            "image classified as %s with %f confidence",
            target.classification,
            target.confidence,
        )
    else:
        logger.warning("no target found")
        sys.exit(-1)

    return target


def calculate_responsibility(
    data: Data, args: CausalArgs, prediction_func, keep_all_maps=False
) -> tuple[ResponsibilityMaps, dict]:
    """Calculates ResponsibilityMaps for input data using given args.

    Runs :py:func:`~rex_xai.responsibility.causal_explanation` for ``args.iters`` iterations,
    and returns a ResponsibilityMaps object and a dictionary containing some statistics about
    the calculation process.
    The ResponsibilityMaps object by default only includes the responsibility map that matches
    the classification of the input data. Set ``keep_all_maps`` to ``True`` to retain all maps.

    Args:
        data: processed input data object
        args: configuration values for ReX
        prediction_func: prediction function for the model
        keep_all_maps: whether to retain all :py:class:`~rex_xai.resp_maps.ResponsibilityMaps`,
            or just the one that matches the target classification.

    Returns:
        tuple containing

        - ResponsibilityMaps: ResponsibilityMaps for the given data, prediction function, and args.
        - dict: statistics for the call of this function that generated the ResponsibilityMaps object
    """

    if data.target is None or data.target.classification is None:
        raise ValueError(
            "No target classification found. Please run `predict_target` before running `calculate_responsibility`."
        )

    maps = ResponsibilityMaps()
    if data.model_height is not None:
        maps.new_map(
            data.target.classification,
            data.model_height,
            data.model_width,
            data.model_depth,
        )
    else:
        maps.new_map(data.target.classification, data.model_height, data.model_width)

    total_passing: int = 0
    total_failing: int = 0
    max_depth_reached: int = 0
    avg_box_size: float = 0.0

    if args.iters > 0:
        for i in trange(args.iters, disable=not args.progress_bar):
            (
                local_maps,
                passing,
                failing,
                depth_reached,
                avg_bs,  # average box size
            ) = causal_explanation(
                i + 1,
                data,
                args,
                prediction_func,
                current_map=maps.get(data.target.classification),
            )

            total_passing += passing
            total_failing += failing
            max_depth_reached = max(max_depth_reached, depth_reached)
            avg_box_size += avg_bs
            # TODO this needs to be smarter. If we only have shallow penetration, then
            # this is doing us a disservice. Perhaps leave merging until after completion
            # of all iterations. Might potentially use a lot of memeory though
            if depth_reached > 1:
                maps.merge(local_maps)

    avg_box_size /= args.iters

    logger.info(
        "Total Passing: %d, Total Failing %d, Max Depth Reached %d, Avg Box Size %f",
        total_passing,
        total_failing,
        max_depth_reached,
        avg_box_size,
    )

    run_stats = {
        "total_passing": total_passing,
        "total_failing": total_failing,
        "max_depth_reached": max_depth_reached,
        "avg_box_size": avg_box_size,
    }

    return maps, run_stats


def analyze(exp: Explanation, data_mode: str | None):
    """Analyzes an Explanation.

    Analyzes the area ratio, entropy difference, insertion and deletion curves for an
    Explanation object, prints them, and returns them.

    Args:
        exp: Explanation object as returned by :py:func:`~rex_xai.explanation._explanation`
        data_mode: Mode of the input data. Entropy difference is only calculated if ``data_mode``
            is one of ["RGB", "L"].

    Returns:
        tuple containing

        - area (float)
        - entropy_diff (float)
        - insertion_curve (float)
        - deletion_curve (float)

    """
    eval = Evaluation(exp)
    rat = eval.ratio()
    ent = None
    max_ent = None
    if data_mode in ("RGB", "RGBA", "L"):
        be, ae = eval.entropy_loss()  # type: ignore
        ent = be - ae
    elif data_mode in ("spectral", "tabular"):
        ent, max_ent = eval.spectral_entropy()

    iauc, dauc = eval.insertion_deletion_curve(
        exp.prediction_func, normalise=exp.args.normalise_curves
    )

    analysis_results = {
        "area": rat,
        "entropy": ent,
        "max_entropy": max_ent,
        "insertion_curve": iauc,
        "deletion_curve": dauc,
    }

    return analysis_results


def _explanation(
    args: CausalArgs,
    model_shape: Tuple[int],
    prediction_func,
    device: tt.device,
    db: Session | None = None,
    path=None,
):
    """Takes a CausalArgs object and model information and returns a Explanation.

    Takes a CausalArgs object, model shape and prediction function and returns an Explanation.
    Depending on the input ``args``, optionally produces output plots, analyses the output
    explanation, and/or writes results to a database.

    Args:
        args: configuration values for ReX
        model_shape: shape of the input tensor of the model, as returned by :py:func:`~rex_xai.explanation.get_prediction_func_from_args()`
        prediction_func: as returned by :py:func:`~rex_xai.explanation.get_prediction_func_from_args()`
        device: as returned by :py:func:`~rex_xai._utils.get_device()`
        db: None or as returned by :py:func:`~rex_xai.database.initialise_rex_db()`

    Returns:
        Explanation:
            An :py:class:`~rex_xai.extraction.Explanation` object containing the causal reponsibility explanation
            calculated using the given ``args``.

    """
    data = load_and_preprocess_data(model_shape, device, args)
    data.set_mask_value(args.mask_value)
    logger.debug(
        "args.mask_value is %s, data.mask_value is %s", args.mask_value, data.mask_value
    )

    data.target = predict_target(data, prediction_func)

    start = time.time()

    logger.info("Calculating responsibility map")
    resp_object, run_stats = calculate_responsibility(data, args, prediction_func)

    logger.info("Extracting explanation from responsibility map")
    clauses = None
    if args.strategy in (Strategy.MultiSpotlight, Strategy.Contrastive):
        exp = MultiExplanation(resp_object, prediction_func, data, args, run_stats)
        exp.extract()

        if args.strategy == Strategy.Contrastive and args.permitted_overlap != 1.0:
            logger.warning(
                "contrastive explanations require a permitted overlap of 1.0, so setting this now"
            )
            args.permitted_overlap = 1.0

        clauses = exp.separate_by(args.permitted_overlap)
        logger.info(f"found the following sets of explanations {clauses}")

        if args.strategy == Strategy.Contrastive:
            clauses = exp.contrastive(clauses)
            args.multi_style = "contrastive"
    else:
        exp = Explanation(resp_object, prediction_func, data, args, run_stats)
        exp.extract(args.strategy)

    if args.analyze:
        logger.info("Analysing explanation")
        results = analyze(exp, data.mode)
        if data.mode == "spectral":
            print(
                f"INFO:ReX:classification {exp.data.target.classification}, area {results['area']}, responsibility entropy {results['entropy']},",  # type: ignore
                f"max entropy {results['max_entropy']}",
                f"insertion curve {results['insertion_curve']}, deletion curve {results['deletion_curve']}",
            )
        else:
            print(
                f"INFO:ReX:classification {exp.data.target.classification}, area {results['area']}, entropy {results['entropy']},",  # type: ignore
                f"insertion curve {results['insertion_curve']}, deletion curve {results['deletion_curve']}",
            )

    end = time.time()
    time_taken = end - start
    logger.info(f"Time taken: {time_taken:.2f}s")

    if args.surface is not None:
        if path is not None:
            pass
        elif args.surface == "show":
            path = None
        else:
            path = args.surface
        exp.surface_plot(path)

    if args.heatmap is not None:
        if args.heatmap == "show":
            path = None
        else:
            path = args.heatmap
        exp.heatmap_plot(path)

    if args.output is not None:
        if path is None:
            if args.output == "show":
                path = None
            else:
                path = args.output
        exp.save(path, clauses=clauses)

    if db is not None:
        if args.strategy == Strategy.MultiSpotlight:
            logger.info("writing multiple explanations to database")
            update_database(
                db,
                exp,
                time_taken,
                multi=True,
            )
        else:
            logger.info("writing to database")
            update_database(
                db,
                exp,
                time_taken,
            )

    return exp


def get_prediction_func_from_args(args: CausalArgs):
    """Takes a CausalArgs object and gets the prediction function and model shape.

    If ``args.script`` specifies a prediction function and model shape, returns these.
    Otherwise gets the prediction function and model shape from the provided model
    file.

    Args:
        args: configuration values for ReX

    Returns:
        tuple containing

        - ``prediction_func``
        - ``model_shape``

    Raises:
        RuntimeError: if an onnx inference instance cannot be created from the provided model file.

    """
    if hasattr(args.script, "prediction_function") and hasattr(
        args.script, "model_shape"
    ):
        prediction_func = args.script.prediction_function  # type: ignore
        model_shape = args.script.model_shape()  # type: ignore
    else:
        ps = get_prediction_function(args)
        if ps is None:
            raise RuntimeError("Unable to create an onnx inference instance")
        else:
            prediction_func, model_shape = ps

    return prediction_func, model_shape


def explanation(
    args: CausalArgs, device: tt.device, db: Session | None = None
) -> Union[Explanation, List[Explanation]]:
    """Takes a CausalArgs object and returns a Explanation.

    Takes a CausalArgs object and returns either an Explanation, or a list of Explanations
    if the input ``args.path`` is a directory rather than a path to a single file.

    Args:
        args: configuration values for ReX
        device: as returned by :py:func:`~rex_xai._utils.get_device()`
        db: None or as returned by :py:func:`~rex_xai.database.initialise_rex_db()`

    Returns:
        Explanation:
            An :py:class:`~rex_xai.extraction.Explanation` object containing the causal reponsibility explanation
            calculated using the given ``args``.

    """

    prediction_func, model_shape = get_prediction_func_from_args(args)

    if isinstance(model_shape[0], int) and model_shape[0] < args.batch_size:
        logger.warning(
            f"Resetting batch size to size of model's first axis: {model_shape[0]}"
        )
        args.batch_size = model_shape[0]

    # directory of data to process
    if os.path.isdir(args.path):
        explanations = []
        dir = args.path
        path = None
        for dir, _, files in os.walk(args.path):
            for f in files:
                to_process = os.path.join(dir, f)
                logger.info("processing %s", to_process)
                current_args = copy.copy(args)
                current_args.path = to_process
                if args.output is not None and args.output != "show":
                    out_dir = os.path.dirname(args.output)
                    name, ext = os.path.splitext(args.output)
                    fname, _ = os.path.splitext(f)
                    path = f"{out_dir}_{fname}_{name}{ext}"
                exp = _explanation(
                    current_args,
                    model_shape,
                    prediction_func,
                    device,
                    db,
                    path=path,
                )
                explanations.append(exp)
        return explanations

    else:
        # a single explanation
        return _explanation(args, model_shape, prediction_func, device, db)



#!/usr/bin/env python
"""simple wrapper around python logging module"""

import logging
import sys
import os


def set_log_level(i: int, rex_logger):
    """sets the logging level for rex"""
    if i == 0:
        rex_logger.setLevel(logging.CRITICAL)
    elif i == 1:
        rex_logger.setLevel(logging.WARNING)
    elif i == 2:
        rex_logger.setLevel(logging.INFO)
    else:
        rex_logger.setLevel(logging.DEBUG)


os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
logger = logging.getLogger("ReX")
logging.basicConfig(stream=sys.stdout, level=logging.WARNING)



#!/usr/bin/env python

"""onnx model management"""

from typing import Optional, Union, List
import sys
import os
import torch as tt
import platform
from scipy.special import softmax
import numpy as np

import onnxruntime as ort
from onnxruntime import InferenceSession
from rex_xai.prediction import Prediction, from_pytorch_tensor
from rex_xai.input_data import Setup

from rex_xai.logger import logger


class OnnxRunner:
    def __init__(self, session: InferenceSession, setup: Setup, device) -> None:
        self.session = session
        self.input_shape = session.get_inputs()[0].shape
        self.output_shape = session.get_outputs()[0].shape
        self.input_name = session.get_inputs()[0].name
        self.output_name = session.get_outputs()[0].name
        self.setup: Setup = setup
        self.device: str = device

    def run_on_cpu(
        self,
        tensors: Union[tt.Tensor, List[tt.Tensor]],
        target: Optional[Prediction],
        raw: bool,
        binary_threshold: Optional[float] = None,
    ):
        """Convert a pytorch tensor, or list of tensors, to numpy arrays on the cpu for onnx inference."""
        # check if it's a single tensor or a list of tensors
        if isinstance(tensors, list):
            tensor_size = tensors[0].shape[0]
        else:
            tensor_size = tensors.shape[0]

        if tensor_size == 1:
            tensors = tensors.detach().cpu().numpy()  # type: ignore
        else:
            tensors = np.stack([t.detach().cpu().numpy() for t in tensors])  # type: ignore

        preds = []

        try:
            prediction = self.session.run(None, {self.input_name: tensors})[0]
            for i in range(0, prediction.shape[0]):
                confidences = softmax(prediction[i])
                if raw:
                    for i in range(len(self.output_shape) - len(confidences.shape)):
                        confidences = np.expand_dims(confidences, axis=0)
                    return confidences
                if binary_threshold is not None:
                    if confidences[0] >= binary_threshold:
                        classification = 1
                    else:
                        classification = 0
                    tc = confidences[0]
                else:
                    classification = np.argmax(confidences)
                    if target is not None:
                        tc = confidences[target.classification]
                    else:
                        tc = None
                preds.append(
                    Prediction(
                        classification,
                        confidences[classification],
                        None,
                        target=target,
                        target_confidence=tc,
                    )
                )

            return preds
        except Exception as e:
            logger.fatal(e)
            sys.exit(-1)

    def run_with_data_on_device(
        self,
        tensors,
        device,
        tsize,
        binary_threshold,
        raw=False,
        device_id=0,
        target=None,
    ):
        # input_shape = self.session.get_inputs()[0].shape # Gets the shape of the input (e.g [batch_size, 3, 224, 224])
        batch_size = len(tensors) if isinstance(tensors, list) else tensors.shape[0]

        if isinstance(tensors, list):
            tensors = [m.contiguous() for m in tensors]
            shape = tuple(
                [batch_size] + list(self.input_shape)[1:]
            )  # batch_size + remaining input shape
            ptr = tensors[0].data_ptr()

        else:
            tensors = tensors.contiguous()
            shape = tuple(tensors.shape)
            ptr = tensors.data_ptr()

        binding = self.session.io_binding()
        binding.bind_input(
            name=self.input_name,
            device_type=device,
            device_id=device_id,
            element_type=np.float32,
            shape=shape,
            buffer_ptr=ptr,
        )

        output_shape = [batch_size] + list(self.output_shape[1:])

        z_tensor = tt.empty(output_shape, dtype=tt.float32, device=device).contiguous()

        binding.bind_output(
            name=self.output_name,
            device_type=device,
            device_id=device_id,
            element_type=np.float32,
            shape=tuple(z_tensor.shape),
            buffer_ptr=z_tensor.data_ptr(),
        )

        self.session.run_with_iobinding(binding)
        if raw:
            return z_tensor
        return from_pytorch_tensor(z_tensor, target=target)

    def gen_prediction_function(self):
        if self.device == "cpu" or self.setup == Setup.ONNXMPS:
            return (
                lambda tensor,
                target=None,
                raw=False,
                binary_threshold=None: self.run_on_cpu(
                    tensor, target, raw, binary_threshold
                ),
                self.input_shape,
            )
        if self.device == "cuda":
            return (
                lambda tensor,
                target=None,
                device=self.device,
                raw=False,
                binary_threshold=None: self.run_with_data_on_device(
                    tensor,
                    device,
                    len(tensor),
                    binary_threshold,
                    raw=raw,
                    target=target,
                ),
                self.input_shape,
            )


def get_prediction_function(args):
    # def get_prediction_function(model_path, gpu: bool, logger_level=3):
    sess_options = ort.SessionOptions()

    ort.set_default_logger_severity(args.ort_logger)

    # are we (trying to) run on the gpu?
    if args.gpu:
        logger.info("using gpu for onnx inference session")
        if platform.uname().system == "Darwin":
            # note this is only true for M+ chips
            providers = ["CoreMLExecutionProvider", "CPUExecutionProvider"]
            sess_options.intra_op_num_threads = args.intra_op_num_threads
            sess_options.inter_op_num_threads = args.inter_op_num_threads
            device = "mps"
            _, ext = os.path.splitext(os.path.basename(args.model))
            # for the moment, onnx does not seem to support data copying on mps, so we fall back to
            # copying data to the cpu for inference
            setup = Setup.ONNXMPS
        else:
            providers = ["CUDAExecutionProvider", "CPUExecutionProvider"]
            device = "cuda"
            setup = Setup.PYTORCH

        # set up sesson with gpu providers
        sess = ort.InferenceSession(
            args.model, sess_options=sess_options, providers=providers
        )  # type: ignore
    # are we running on cpu?
    else:
        logger.info("using cpu for onnx inference session")
        providers = ["CPUExecutionProvider"]
        sess = ort.InferenceSession(
            args.model, sess_options=sess_options, providers=providers
        )
        device = "cpu"
        setup = Setup.PYTORCH

    onnx_session = OnnxRunner(sess, setup, device)

    return onnx_session.gen_prediction_function()



#!/usr/bin/env python
from typing import Optional
import numpy as np
import torch as tt

from enum import Enum

from rex_xai.occlusions import spectral_occlusion, context_occlusion
from rex_xai.prediction import Prediction
from rex_xai.logger import logger
from rex_xai._utils import ReXDataError

Setup = Enum("Setup", ["ONNXMPS", "ONNX", "PYTORCH"])


def _guess_mode(input):
    if hasattr(input, "mode"):
        return input.mode
    if hasattr(input, "shape"):
        if len(input.shape) == 4:
            return "voxel"
        else:
            return "spectral"


class Data:
    def __init__(self, input, model_shape, device, mode=None, process=True) -> None:
        self.input = input
        self.mode = None
        self.target: Optional[Prediction] = None
        self.device = device
        self.setup: Optional[Setup] = None

        self.mode = mode
        if mode is None:
            self.mode = _guess_mode(input)

        self.model_shape = model_shape
        height, width, channels, order, depth = self.__get_shape()
        self.model_height: Optional[int] = height
        self.model_width: Optional[int] = width
        self.model_depth: Optional[int] = depth
        self.model_channels: Optional[int] = channels if channels is not None else 1
        self.model_order = order
        self.mask_value = None
        self.background = None
        self.context = None

        if process:
            # RGB model but greyscale input so we convert greyscale to pseudo-RGB
            if self.model_channels == 3 and self.mode == "L":
                self.input = self.input.convert("RGB")
                self.mode = "RGB"
            elif self.mode == "RGB" and self.model_order == "first":
                self.transposed = True
            elif self.mode in ("tabular", "spectral"):
                self.data = self.input
                self.match_data_to_model_shape()
            elif self.mode == "voxel":
                self.data = self.input
            else:
                raise NotImplementedError

    def set_height(self, h: int):
        self.model_height = h

    def set_width(self, w: int):
        self.model_width = w

    def set_channels(self, c=None):
        self.model_channels = c

    def __repr__(self) -> str:
        data_info = f"Data: {self.mode}, {self.model_shape}, {self.model_height}, {self.model_width}, {self.model_channels}, {self.model_order}"
        if self.target is not None:
            target_info = repr(self.target)
            data_info = data_info + "\n\t Target:" + target_info
        return data_info

    def set_classification(self, cl):
        self.classification = cl

    def match_data_to_model_shape(self):
        """
        a PIL image has the from H * W * C, so
        if the model takes C * H * W we need to transpose self.data to
        get it into the correct form for the model to consume
        This function does *not* add in the batch channel at the beginning
        """
        assert self.data is not None
        if self.mode == "RGB" and self.model_order == "first":
            self.data = self.data.transpose(2, 0, 1)  # type: ignore
            self.transposed = True
        if self.mode == "L" and self.model_order == "first":
            self.data = self.data.transpose(1, 0)
            self.transposed = True
        if self.mode == "L" and self.model_order == "last":
            self.data = self.data.transpose(1, 0)
        if self.mode == "tabular" or self.mode == "spectral":
            self.data = self.generic_tab_preprocess()
        if self.mode == "voxel":
            pass
        self.data = self.try_unsqueeze()

    def generic_tab_preprocess(self):
        if isinstance(self.input, np.ndarray):
            self.data = self.input.astype("float32")
            arr = tt.from_numpy(self.data).to(self.device)
        else:
            arr = self.input
        for _ in range(len(self.model_shape) - len(arr.shape)):
            arr = arr.unsqueeze(0)
        return arr

    def load_data(self, astype="float32"):
        img = self.input.resize((self.model_height, self.model_width))
        img = np.array(img).astype(astype)
        self.data = img
        self.match_data_to_model_shape()
        self.data = tt.from_numpy(self.data).to(self.device)

    def _normalise(self, means, stds, astype, norm):
        assert self.data is not None

        normed_data = self.data
        if norm is not None:
            normed_data /= norm

        if self.model_order == "first" and self.model_channels == 3:
            if means is not None:
                for i, m in enumerate(means):
                    normed_data[:, i, :, :] = normed_data[:, i, :, :] - m
            if stds is not None:
                for i, s in enumerate(stds):
                    normed_data[:, i, :, :] = normed_data[:, i, :, :] / s

        if self.model_order == "last" and self.model_channels == 3:
            if means is not None:
                for i, m in enumerate(means):
                    normed_data[:, :, i] = normed_data[:, :, i] - m
            if stds is not None:
                for i, s in enumerate(stds):
                    normed_data[:, :, i] = normed_data[:, :, i] / s

        # greyscale
        if self.model_channels == 1:
            if means is not None:
                for i, m in enumerate(means):
                    normed_data[i] = normed_data[i] - m
            if stds is not None:
                for i, s in enumerate(stds):
                    normed_data[i] = normed_data[i] / s

        return normed_data

    def try_unsqueeze(self):
        out = self.data
        if self.model_order == "first":
            dim = 0
        else:
            dim = -1
        if isinstance(self.data, tt.Tensor):
            for _ in range(len(self.model_shape) - len(self.data.shape) - 1):
                out = tt.unsqueeze(out, dim=dim)  # type: ignore
            out = tt.unsqueeze(out, dim=0)  # type: ignore
        else:
            for _ in range(len(self.model_shape) - len(self.data.shape) - 1):  # type: ignore
                out = np.expand_dims(out, axis=dim)  # type: ignore
            out = np.expand_dims(out, axis=0)  # type: ignore
        return out

    def generic_image_preprocess(
        self,
        means=None,
        stds=None,
        astype="float32",
        norm: Optional[float] = 255.0,
    ):
        self.load_data(astype=astype)

        if self.mode == "RGB" and self.data is not None:
            self.data = self._normalise(means, stds, astype, norm)
            self.try_unsqueeze()
        if self.mode == "L":
            self.data = self._normalise(means, stds, astype, norm)

    def __get_shape(self):
        """returns height, width, channels, order, depth for the model"""
        if self.mode == "spectral":
            # an array of the form (h, w), so no channel info or order or depth
            if len(self.model_shape) == 2:
                return self.model_shape[0], self.model_shape[1], 1, None, None
            # an array of the form (batch, h, w), so no channel info or order or depth
            if len(self.model_shape) == 3:
                return self.model_shape[1], self.model_shape[2], 1, None, None
        if self.mode in ("RGB", "RGBA"):
            if len(self.model_shape) == 4:
                _, a, b, c = self.model_shape
                if a in (1, 3, 4):
                    return b, c, a, "first", None
                else:
                    return a, b, c, "last", None
        if self.mode == "voxel":
            if len(self.model_shape) == 4:
                batch, w, h, d = self.model_shape  # If batch is present
                return w, h, None, None, d
            else:
                w, h, d = self.model_shape
                return w, h, None, None, d

        raise ReXDataError(
            f"Incompatible 'mode' {self.mode}  and 'model_shape' ({self.model_shape}), cannot get valid shape of Data object so returning None"
        )

        # elif self.mode in ("RGB", "RGBA", "L") and len(self.model_shape) == 4:
        #     _, a, b, c = self.model_shape
        #     if a == 1 or a == 3 or a == 4:
        #         return b, c, a, "first"
        #     else:
        #         return a, b, c, "last"
        # elif self.mode == "voxel":
        #     pass
        # else:
        #     logger.warning("Incompatible 'mode' and 'model_shape', cannot get valid shape of Data object so returning None")
        #     return None, None, None, None

    def set_mask_value(self, m):
        assert self.data is not None
        # if m is a number, then if might still need to be normalised

        if m == "spectral" and self.mode != "spectral":
            logger.warning(
                "Mask value 'spectral' can only be used if mode is also 'spectral', using default mask value 0 instead"
            )
            m = 0

        match m:
            case int() | float() as m:
                self.mask_value = m
            case "min":
                self.mask_value = tt.min(self.data).item()  # type: ignore
            case "mean":
                self.mask_value = tt.mean(self.data).item()  # type: ignore
            case "spectral":
                self.mask_value = lambda m, d: spectral_occlusion(
                    m, d, device=self.device
                )
            case "context":
                self.mask_value = lambda m, d: context_occlusion(m, d, self.context)
                # TODO: Add args for noise and setting the context as currently only available through custom script
            case _:
                raise ValueError(
                    f"Invalid mask value {m}. Should be an integer, float, or one of 'min', 'mean', 'spectral'"
                )



#!/usr/bin/env python

"""
calculate causal responsibility
"""

from collections import deque
from typing import List

import numpy as np
import torch as tt

try:
    from anytree.cachedsearch import find
except ImportError:
    from anytree.search import find


from rex_xai.box import average_box_size, initialise_tree
from rex_xai.config import CausalArgs, Queue
from rex_xai.input_data import Data
from rex_xai.logger import logger
from rex_xai.mutant import Mutant, get_combinations, _apply_to_data
from rex_xai.resp_maps import ResponsibilityMaps
from rex_xai.prediction import Prediction


def subbox(tree, name, max_depth, min_size, mode, r_map=None):
    """Split a box into 4 contiguous children or None if no possible.

    @param tree: a search tree
    @param name: a node name to search for in <tree>
    @param max_depth: tree depth limit for new children
    @param min_size: minimum new child size
    @param mode: spectral, tabular, RGB, L or voxel
    @param r_map=None: responsibility map

    @return None or new children
    """
    active = find(tree, lambda node: node.name == name)
    if active is not None:
        if active.depth <= max_depth and len(active.children) == 0:
            active.add_children_to_tree(min_size, mode, r_map)
        return active.children
    return None


def update_depth_reached(depth_reached, passing):
    """Update the max tree depth reached.

    @param depth_reached: current max_depth
    @param passing: a list of passing Mutant

    @return int
    """
    mp = [m.depth for m in passing]
    if mp == []:
        mp = 0
    else:
        mp = max(mp)
    return max(depth_reached, mp)


def prune(mutants: List[Mutant], technique=Queue.Intersection, keep=None):
    """Reduce the search queue to improve both efficiency and (possibly) result quality.

    @param mutants: a list of passing Mutant objects
    @param technique: a Queue enum
    @param keep=None: how many items to keep in the queue, all if keep is None

    @return a list of mutants of length <= keep
    """
    # we use "none" when we are looking for multiple explanations. It has a tendency
    # to produce flatter landscapes than intersection
    if technique == Queue.All:
        return mutants
    if technique == Queue.Intersection and len(mutants) > 1:
        inter = set()
        i = 0
        while len(inter) == 0 and i < len(mutants):
            names = [mutant.get_active_boxes() for mutant in mutants]
            head = set(names[i])
            tail = names[i + 1 :]
            inter = head.intersection(*tail)
            i += 1
        mutants = [m for m in mutants if inter <= set(m.get_active_boxes())]
        ordered = sorted(mutants, key=lambda x: x.area())
        if keep is not None:
            return ordered[:keep]
        return ordered

    if technique == Queue.Area:
        ordered = sorted(mutants, key=lambda x: x.area())
        if keep is not None:
            return ordered[:keep]
        return ordered

    return mutants


def causal_explanation(
    process, data: Data, args: CausalArgs, prediction_func, current_map=None
):
    """Calculate causal responsiblity.

    @param process: an integer value
    @param data: a Data object
    @param args: a CausalArgs object
    @param prediction_func: a higher order
        function that calls a model and return a Prediction object
    """

    assert data.target is not None

    if args.seed is not None:
        np.random.seed(args.seed + process)
        tt.manual_seed(args.seed + process)

    search_tree = initialise_tree(
        data.model_height,
        data.model_width,
        args.distribution,
        args.distribution_args,
        d_lim=data.model_depth,
    )

    total_work = 0
    total_passing = 0
    total_failing = 0

    depth_reached = 0

    # The <queue> is a list of strings in the form "R:x:y:...n"
    queue = deque(search_tree.name)

    local_maps = ResponsibilityMaps()

    # a <job> is of the form "R:x:y:...n", where x,y...n are integers.
    # This is both the unique name for a passing mutant and the node name for
    # the node in <search_tree>
    while True:
        passing = []
        while len(queue) != 0:
            job = queue.popleft()
            sub_jobs = job.split("_")

            todo = len(sub_jobs) - 1
            for ai, active in enumerate(sub_jobs):
                static = [sj for x, sj in enumerate(sub_jobs) if x != ai]

                child_boxes = subbox(
                    search_tree,
                    active,
                    args.tree_depth,
                    args.min_box_size,
                    data.mode,
                    r_map=current_map,
                )

                if child_boxes is None or len(child_boxes) == 0:
                    logger.debug("no children, breaking")
                    break

                mutants = np.empty(14, dtype=np.object_)
                if child_boxes is not None:
                    for j, combination in enumerate(get_combinations()):
                        nps = [child_boxes[i] for i in combination]
                        current = "_".join([b.name for b in nps])

                        m = Mutant(
                            data,
                            static=static,
                            active=current,
                            masking_func=data.mask_value,
                        )
                        m.set_active_mask_regions(nps)
                        m.set_static_mask_regions(static, search_tree)
                        mutants[j] = m

                work_done = len(mutants)

                def apply_mask(m):
                    if args.mask_value == "context":
                        return _apply_to_data(m.mask, data, data.mask_value)
                    return tt.where(m.mask, data.data, data.mask_value)

                if data.mode in ("spectral", "tabular"):
                    preds: List[Prediction] = [
                        prediction_func(_apply_to_data(m.mask, data, data.mask_value))[
                            0
                        ]
                        for m in mutants
                    ]
                else:
                    # TODO this needs testing
                    if args.batch_size == 1:
                        preds = [
                            prediction_func(
                                apply_mask(m),  #  type: ignore
                                data.target,
                                binary_threshold=args.binary_threshold,
                            )[0]
                            for m in mutants
                        ]  # type: ignore
                    else:
                        tensors = tt.stack(
                            [
                                apply_mask(m)  #  type: ignore
                                for m in mutants
                            ]
                        )  # type: ignore
                        if len(tensors.shape) > len(data.model_shape):
                            tensors = tensors.squeeze(1)
                        preds: List[Prediction] = prediction_func(
                            tensors,
                            data.target,
                            binary_threshold=args.binary_threshold,
                        )

                for i, m in enumerate(mutants):
                    m.prediction = preds[i]
                    m.update_status(data.target)

                passing: List[Mutant] = list(
                    filter(
                        lambda m: m.passing
                        and m.prediction.confidence
                        >= (data.target.confidence * args.confidence_filter),  # type: ignore
                        mutants,
                    )
                )

                if args.verbosity > 3:
                    n = 0
                    for m in mutants:
                        m.save_mutant(
                            data,
                            f"{process}_{m.depth}_{n}_{m.prediction.confidence}_{m.passing}.png",
                        )
                        n += 1

                total_passing += len(passing)
                total_failing += work_done - len(passing)
                total_work += work_done

                # we have no passing occlusions
                if not passing:
                    if ai == todo:
                        logger.debug(
                            "there are no passing mutants at %d, so quitting here",
                            depth_reached,
                        )
                        # logger.debug(global_queue)
                        break

                # something passed...
                else:
                    # update responsibilities
                    local_maps.update_maps(mutants, args, data, search_tree)  # type: ignore

                    # reduce the elements to add to the search queue
                    passing = prune(
                        passing, technique=args.queue_style, keep=args.queue_len
                    )  # type: ignore

                depth_reached = update_depth_reached(depth_reached, passing)

        # if we are too deep into the tree, break from the loop
        if depth_reached > args.tree_depth and ai == todo:  # type: ignore
            logger.info("breaking at %s as max tree depth reached", depth_reached)
            break

        if args.search_limit is not None and total_work > args.search_limit:
            logger.info("exceeded total work limit for this iteration")
            break

        if args.queue_style == Queue.DC:
            update = list(set([m.get_name() for m in passing] + list(queue)))
        else:
            update = [m.get_name() for m in passing]
        if update == []:
            logger.debug("nothing left in the queue")
            break

        # our new search queue, which takes us back to the beginning
        queue = deque(update)

    # clear up unneeded mutants and boxes
    if data.device == "mps":
        with tt.no_grad():
            tt.mps.empty_cache()
    elif data.device == "cuda":
        with tt.no_grad():
            tt.cuda.empty_cache()

    logger.info(
        "total work %d with %d passing and %d failing, max depth explored %d",
        total_work,
        total_passing,
        total_failing,
        depth_reached,
    )

    return (
        local_maps,
        total_passing,
        total_failing,
        depth_reached,
        average_box_size(search_tree, depth_reached),
    )



#!/usr/bin/env python

"""generate boxes which together create a occlusion.
This occlusion is realised in the form of a mask over an image"""

from __future__ import annotations

from typing import List, Tuple, Dict
from anytree import LevelOrderGroupIter, NodeMixin, RenderTree
import numpy as np

from rex_xai.distributions import Distribution, random_coords
from rex_xai.logger import logger

# Enums for different axes
class Axes:
    ROW = 0
    COL = 1
    DEPTH = 2


class BoxInternal:
    """a box is part of an occulsion, a collection of boxes which form a mask over an image"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        row_start,
        row_stop,
        col_start,
        col_stop,
        depth_start=None,  # Optional depth_start and depth_stop for 3D data
        depth_stop=None,
        distribution=None,
        distribution_args=None,
        name="",
    ) -> None:
        self.name = name
        self.distribution = distribution
        self.distribution_args = distribution_args
        self.row_start: int = row_start
        self.row_stop: int = row_stop
        self.col_start: int = col_start
        self.col_stop: int = col_stop
        self.depth_start = depth_start
        self.depth_stop = depth_stop

    def __repr__(self) -> str:
        if self.depth_start is not None and self.depth_stop is not None:
            return (
                f"Box < name: {self.name}, row_start: {self.row_start}, "
                + f"row_stop: {self.row_stop}, col_start: {self.col_start}, "
                + f"col_stop: {self.col_stop}, depth_start: {self.depth_start}, "
                + f"depth_stop: {self.depth_stop}, volume: {self.area()}"
            )
        else:
            return (
                f"Box < name: {self.name}, row_start: {self.row_start}, "
                + f"row_stop: {self.row_stop}, col_start: {self.col_start}, "
                + f"col_stop: {self.col_stop}, area: {self.area()}"
            )

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other) -> bool:
        if isinstance(other, int):
            return self.name == other
        if isinstance(other, type(self)):
            return NotImplemented
        return self.name == other.name

    def update_name(self, name: str):
        """sets the name of a box"""
        self.name += name

    def shape(self):
        """returns (width, height) of a box if 2d data, else returns (width, height, depth)"""
        if self.depth_start is not None and self.depth_stop is not None:
            return (
                self.row_stop - self.row_start,
                self.col_stop - self.col_start,
                self.depth_stop - self.depth_start,
            )
        else:
            return (self.row_stop - self.row_start, self.col_stop - self.col_start)

    def corners(self):
        """Return (Wstart, Wstop, Hstart, Hstop) of current box if 2d data, else returns (Wstart, Wstop, Hstart, Hstop. Dstart, Dstop)"""
        if self.depth_start is not None and self.depth_stop is not None:
            return (
                self.row_start,
                self.row_stop,
                self.col_start,
                self.col_stop,
                self.depth_start,
                self.depth_stop,
            )
        else:
            return (self.row_start, self.row_stop, self.col_start, self.col_stop)

    def __1d_parts(self):
        c1 = random_coords(self.distribution, [self.col_stop - self.col_start])
        if c1 is not None and isinstance(c1, np.ndarray):
            c1 = c1[0] + self.col_start

        c2 = random_coords(self.distribution, [self.col_stop - self.col_start])
        if c2 is not None and isinstance(c2, np.ndarray):
            c2 = c2[0] + self.col_start

        c3 = random_coords(self.distribution, [self.col_stop - self.col_start])
        if c3 is not None and isinstance(c3, np.ndarray):
            c3 = c3[0] + self.col_start

        if c1 is None or c2 is None or c3 is None:
            return None

        ordered = sorted([c1, c2, c3])

        b0 = Box(
            0,
            0,
            self.col_start,
            ordered[0],
            distribution=self.distribution,
            distribution_args=self.distribution_args,
            name=self.name,
        )
        b0.update_name(":0")

        b1 = Box(
            0,
            0,
            ordered[0],
            ordered[1],
            distribution=self.distribution,
            distribution_args=self.distribution_args,
            name=self.name,
        )
        b1.update_name(":1")

        b2 = Box(
            0,
            0,
            ordered[1],
            ordered[2],
            distribution=self.distribution,
            distribution_args=self.distribution_args,
            name=self.name,
        )
        b2.update_name(":2")

        b3 = Box(
            0,
            0,
            ordered[2],
            self.col_stop,
            distribution=self.distribution,
            distribution_args=self.distribution_args,
            name=self.name,
        )
        b3.update_name(":3")

        return [b0, b1, b2, b3]

    def __2d_parts(self, map=None):
        if self.distribution == Distribution.Adaptive:
            pos = random_coords(self.distribution, self.corners(), map=map)
        else:
            h = int(self.row_stop - self.row_start)
            w = int(self.col_stop - self.col_start)
            space: int = h * w
            pos = random_coords(
                self.distribution, space, h, w, self.distribution_args, map=map
            )

        if pos is None:
            return
        coords = np.unravel_index(
            pos,
            (self.row_stop - self.row_start, self.col_stop - self.col_start),
        )  # type: ignore
        row: int = int(coords[0]) + self.row_start
        col: int = int(coords[1]) + self.col_start

        b0 = Box(
            self.row_start,
            row,
            self.col_start,
            col,
            distribution=self.distribution,
            distribution_args=self.distribution_args,
            name=self.name,
        )
        b0.update_name(":0")

        b1 = Box(
            self.row_start,
            row,
            col,
            self.col_stop,
            distribution=self.distribution,
            distribution_args=self.distribution_args,
            name=self.name,
        )
        b1.update_name(":1")

        b2 = Box(
            row,
            self.row_stop,
            self.col_start,
            col,
            distribution=self.distribution,
            distribution_args=self.distribution_args,
            name=self.name,
        )
        b2.update_name(":2")

        b3 = Box(
            row,
            self.row_stop,
            col,
            self.col_stop,
            distribution=self.distribution,
            distribution_args=self.distribution_args,
            name=self.name,
        )
        b3.update_name(":3")

        return [b0, b1, b2, b3]

    def __3d_parts(self, map=None):
        """
        Create 4 boxes from the original box passed in as an argument.
        Pick two axes to split on randomly using the Axes class and create boxes.
        """

        # Pick two axes to split on randomly using the axes class
        axes = [Axes.ROW, Axes.COL, Axes.DEPTH]
        selected_axes = np.random.choice(axes, 2, replace=False)
        # Range of the different axes to split on
        ranges: Dict = {
            Axes.ROW: [self.row_start, self.row_stop],
            Axes.COL: [self.col_start, self.col_stop],
            Axes.DEPTH: [self.depth_start, self.depth_stop],
        }
        range1 = ranges[selected_axes[0]]
        range2 = ranges[selected_axes[1]]
        logger.debug(f"Selected axes: {selected_axes}, which have a range of values, {range1} and {range2}")
        # Get the random coordinates for the two axes
        space = range1[1] - range1[0]
        if space == 0 or space == 1:
            c1 = np.random.choice([range1[0], range1[1]])
        else:
            c1 = random_coords(
                self.distribution,
                space,
                range1[0],
                range1[1],
                self.distribution_args,
                map=map,
            )
            c1 = range1[0] + c1
        space = range2[1] - range2[0]
        if space == 0 or space == 1:
            c2 = np.random.choice([range2[0], range2[1]])
        else:
            c2 = random_coords(
                self.distribution,
                space,
                range2[0],
                range2[1],
                self.distribution_args,
                map=map,
            )
            c2 = range2[0] + c2
        logger.debug(f"Random coordinates picked: {c1} and {c2}")
        # If any of the coordinates are None, return None
        if c1 == -1 or c2 == -1:
            return None
        # Create boxes depending on the selected axes
        boxes = []

        subboxes = self.create_box(selected_axes[0], c1, self)
        for i, box in enumerate(subboxes):
            boxes.extend(self.create_box(selected_axes[1], c2, box))
        # Update the names of the boxes
        for i, box in enumerate(boxes):
            box.update_name(f":{i}")
        return boxes

    def create_box(self, axes, c1, box: BoxInternal):
        """
        Create a box depending on the selected axis and the random coordinate
        This will create 2 boxes from the original box passed in as an argument.
        params:
            axes: The axis to split on
            c1: The random coordinate to split on
            box: The original box to split on
        """
        if axes == Axes.ROW:
            b0 = Box(
                box.row_start,
                c1,
                box.col_start,
                box.col_stop,
                box.depth_start,
                box.depth_stop,
                distribution=self.distribution,
                distribution_args=self.distribution_args,
                name=self.name,
            )
            b1 = Box(
                c1,
                box.row_stop,
                box.col_start,
                box.col_stop,
                box.depth_start,
                box.depth_stop,
                distribution=self.distribution,
                distribution_args=self.distribution_args,
                name=self.name,
            )
            return [b0, b1]
        if axes == Axes.COL:
            b0 = Box(
                box.row_start,
                box.row_stop,
                box.col_start,
                c1,
                box.depth_start,
                box.depth_stop,
                distribution=self.distribution,
                distribution_args=self.distribution_args,
                name=self.name,
            )
            b1 = Box(
                box.row_start,
                box.row_stop,
                c1,
                box.col_stop,
                box.depth_start,
                box.depth_stop,
                distribution=self.distribution,
                distribution_args=self.distribution_args,
                name=self.name,
            )
            return [b0, b1]
        if axes == Axes.DEPTH:
            b0 = Box(
                box.row_start,
                box.row_stop,
                box.col_start,
                box.col_stop,
                box.depth_start,
                c1,
                distribution=self.distribution,
                distribution_args=self.distribution_args,
                name=self.name,
            )
            b1 = Box(
                box.row_start,
                box.row_stop,
                box.col_start,
                box.col_stop,
                c1,
                box.depth_stop,
                distribution=self.distribution,
                distribution_args=self.distribution_args,
                name=self.name,
            )
            return [b0, b1]

    def spawn_children(self, min_size, mode, map=None) -> List[Box] | Tuple | None:
        """spawn subboxes from <self>"""
        if self.area() < min_size:
            return ()

        # if we have tabular data, we are of the form (1, x) so we grab 3 random numbers from the
        # distribution and partition with those
        if mode == "tabular" or mode == "spectral":
            return self.__1d_parts()

        # we have a 2 dimension image
        if mode == "RGB" or mode == "L":
            return self.__2d_parts(map=map)

        # we have a 3 dimension data
        if mode == "voxel":
            return self.__3d_parts(map=map)

        return []

    def area(self) -> int:
        """returns the area of a box"""

        if self.depth_start is not None and self.depth_stop is not None:
            return (
                (self.row_stop - self.row_start)
                * (self.col_stop - self.col_start)
                * (self.depth_stop - self.depth_start)
            )
        else:
            if self.row_start == 0 and self.row_stop == 0:
                return self.col_stop - self.col_start
            return (self.row_stop - self.row_start) * (self.col_stop - self.col_start)


class Box(BoxInternal, NodeMixin):
    """An object that represents a box, a set of
    boxes forms an occlusion"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        row_start,
        row_stop,
        col_start,
        col_stop,
        depth_start=None,
        depth_stop=None,
        distribution=None,
        distribution_args=None,
        name="",
        parent=None,
        children=None,
    ) -> None:
        super().__init__(
            row_start,
            row_stop,
            col_start,
            col_stop,
            depth_start,
            depth_stop,
            distribution,
            distribution_args,
            name,
        )
        self.parent = parent
        # anytree nodemixin needs every field to be a tuple, and NoneType object is not iterable,
        # so to get around this limitaton, we set children to be an empty tuple if the box has not spawned any subboxes
        self.children = () if children is None else children

    def add_children_to_tree(self, min_size, mode, map):
        """adds a list of boxes (children) to box tree"""
        if self.children == ():
            try:
                self.children = self.spawn_children(min_size, mode, map)
            except TypeError:
                self.children = ()


def initialise_tree(
    r_lim,
    c_lim,
    distribution,
    distribution_args,
    r_start=0,
    c_start=0,
    d_start=None,
    d_lim=None,
) -> Box:
    """initialise box tree with root node, the whole image"""
    if d_lim is not None and d_start is None:
        d_start = 0
    return Box(
        r_start,
        r_lim,
        c_start,
        c_lim,
        depth_start=d_start,
        depth_stop=d_lim,
        distribution=distribution,
        distribution_args=distribution_args,
        name="R",
    )


def show_tree(tree):
    """Print the box tree to the terminal."""
    print(RenderTree(tree))


def average_box_size(tree, d) -> float:
    """Calculate the average box size at depth <d> of <tree>."""
    areas = [
        [node.area() for node in children] for children in LevelOrderGroupIter(tree)
    ]
    try:
        return np.mean(areas[d], axis=0)
    except IndexError:
        return 0.0


def box_dimensions(
    box: Box,
) -> tuple[int, int, int, int, int, int] | tuple[int, int, int, int]:
    """Returns box dimensions as a 4-tuple or 6-tuple depending on whether the box is 2D or 3D.

    @param box: Box
    @return (int, int, int, int) | (int, int, int, int, int, int)
    """
    if box.depth_start is not None and box.depth_stop is not None:
        return (
            box.row_start,
            box.row_stop,
            box.col_start,
            box.col_stop,
            box.depth_start,
            box.depth_stop,
        )
    else:
        return (box.row_start, box.row_stop, box.col_start, box.col_stop)


def boxes_name_and_dimensions(boxes: List[Box]):
    """Returns a list of all boxes and their dimensions"""
    return [(box.name, box_dimensions(box)) for box in boxes]



#!/usr/bin/env python3

"""config management"""

import argparse
import importlib.util
import os
from os.path import exists, expanduser
from types import ModuleType
from typing import List, Optional, Union

import matplotlib as mpl
import toml  # type: ignore

from rex_xai._utils import (
    Queue,
    ReXError,
    ReXPathError,
    ReXTomlError,
    Strategy,
    version,
)
from rex_xai.distributions import Distribution, str2distribution
from rex_xai.logger import logger


class Args:
    """args argument object"""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-few-public-methods
    def __init__(self) -> None:
        self.config_location: Optional[str] = None
        # input file
        self.path: str = ""
        self.model = None
        self.mode: Optional[str] = None
        self.shape: None = None
        self.db: Optional[str] = None
        # gpu support
        self.gpu: bool = True
        # for reproducability
        self.seed: Union[int, float, None] = None
        # for custom scripts (when used as cmdline tool)
        self.script: Optional[ModuleType] = None
        self.script_location = None
        self.processed = False
        # onnx processing
        self.means = None
        self.stds = None
        self.norm: Optional[float] = 255.0
        self.binary_threshold = None
        self.intra_op_num_threads = 8
        self.inter_op_num_threads = 8
        self.ort_logger = 3
        # verbosity
        self.verbosity = 1
        # whether to show progress bar or not
        self.progress_bar = True
        # save explanation to output
        self.output = None
        self.surface: Optional[str] = None
        self.heatmap: Optional[str] = None
        self.info = True
        self.raw: bool = False
        self.colour: int = 200
        self.mark_segments = False
        self.alpha = 0.2
        self.all = False
        self.resize = False
        self.grid = False
        self.heatmap_colours = "magma"
        self.multi_style = "composite"
        # explanation production strategy
        self.strategy: Strategy = Strategy.Global
        self.chunk_size = 25
        self.minimum_confidence_threshold = 0.0
        self.batch_size: int = 1
        # args for spatial strategy
        self.spatial_initial_radius: int = 25
        self.spatial_radius_eta: float = 0.2
        self.no_expansions = 4
        # spotlight args
        self.spotlights: int = 10
        self.spotlight_size: int = 20
        self.spotlight_eta: float = 0.2
        self.spotlight_step: int = 5
        self.spotlight_objective_function: str = "none"
        self.max_spotlight_budget = 40
        self.permitted_overlap: float = 0.0
        # analysis
        self.analyze: bool = False
        self.insertion_step = 100
        self.normalise_curves = True

    def __repr__(self) -> str:
        return (
            f"Args <file: {self.path}, model: {self.model}, "
            + f"gpu: {self.gpu}, "
            + f"mode: {self.mode}, "
            + f"progress_bar: {self.progress_bar}, "
            + f"output_file: {self.output}, surface_plot: {self.surface}, "
            + f"heatmap_plot: {self.heatmap}, "
            + f"onnx_means: {self.means}, onnx_stds: {self.stds}, onnx_norm: {self.norm} "
            + f"onnx_inter_op_threads: {self.inter_op_num_threads}, onnx_intra_op_threads: {self.intra_op_num_threads}, onnx_logger: {self.ort_logger}"
            + f"explanation_strategy: {self.strategy}, "
            + f"min_confidence_scalar: {self.minimum_confidence_threshold}, "
            + f"chunk size: {self.chunk_size}, "
            + f"spatial_radius: {self.spatial_initial_radius}, "
            + f"spatial_eta: {self.spatial_radius_eta}, seed: {self.seed}, "
            + f"db: {self.db}, "
            + f"script: {self.script_location}, verbosity: {self.verbosity}, "
            + f"spotlights: {self.spotlights}, spotlight_size: {self.spotlight_size}, "
            + f"spotlight_eta: {self.spotlight_eta}, "
            + f"no_expansions: {self.no_expansions}, "
            + f"obj_function: {self.spotlight_objective_function}, "
        )


class CausalArgs(Args):
    """Creates a causal args object"""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-few-public-methods

    def __init__(self) -> None:
        super().__init__()
        self.config_location = None
        self.tree_depth: int = 10
        self.search_limit: Optional[int] = None
        self.mask_value: Union[int, float, str] = 0
        self.confidence_filter = 0.0
        self.min_box_size: int = 10
        self.segmentation = False
        self.data_location: Optional[str] = None
        self.distribution: Distribution = Distribution.Uniform
        self.distribution_args: Optional[List] = None
        self.blend = 0.0
        self.weighted: bool = False
        self.iters = 20
        self.concentrate = False
        # queue management
        self.queue_len = 1
        self.queue_style = Queue.Area

        if self.min_box_size is not None:
            self.chunk_size = self.min_box_size

    def __repr__(self) -> str:
        return (
            "Causal Args <"
            + Args.__repr__(self)
            + f"config_location: {self.config_location}, "
            + f"mask_value: {self.mask_value}, "
            + f"tree_depth: {self.tree_depth}, search_limit: {self.search_limit}, "
            + f"min_box_size: {self.min_box_size}, weighted: {self.weighted}, "
            + f"confidence_filter: {self.confidence_filter}, "
            + f"data_locations: {self.data_location}, distribution: {self.distribution}, "
            + f"distribution_args: {self.distribution_args}, "
            + f"queue_len: {self.queue_len}, queue_style {self.queue_style}, "
            + f"concentrate: {self.concentrate}, "
            + f"iterations: {self.iters}>"
        )


def read_config_file(path):
    if path is None:
        return None
    if not os.path.isfile(path):
        raise ReXPathError(path)
    try:
        file_args = toml.load(path)
        return file_args
    except Exception as e:
        raise ReXTomlError(e)


def rex_ascii():
    return (
        "                       @@@@@@@@@@@@@@@                                  \n"
        + "                       @@  @@@@@@@@@@@                                \n"
        + "                     @@@@  @@@@@@@@@@@@@@@@@                          \n"
        + "                     @@@@@@@@@@@@@@@@@@@@@@@                          \n"
        + "                     @@@@@@@@@@@@@@@@@@@@@@@                          \n"
        + "                     @@@@@@@@@@                                       \n"
        + "                     @@@@@@@@@@@@@@@@         %%%                     \n"
        + "   @@              @@@@@@@@@@@                %%%%                    \n"
        + "   @@              @@@@@@@   %%@               %%%%             %%    \n"
        + "   @@            @@@@@@@@ @%%%%%%%              %%%%           %%     \n"
        + "   @@@@        @@@@@@@@@@ %%%%  %%%%%%           %%%%        %%       \n"
        + "   @@@@@@    @@@@@@@@@@@@ %%%%     %%%%%%      %   %%%     %%         \n"
        + "   @@@@@@@@@@@@@@@@@@@@@@ %%%%       %      %%      %%%   %           \n"
        + "     @@@@@@@@@@@@@@@@@@@@ %%%%    %%      %          %%%%             \n"
        + "       @@@@@@@@@@@@@@@@@@ %%%   %%     %%%%%%%%%%%    %%%             \n"
        + "         @@@@@@@@@@@@@@@@ %%%  %%%%   %%%%           % %%%            \n"
        + "         @@@@@@@@@@@@@@@  %%%   %%%%   %%%%        %@   %%%           \n"
        + "           @@@@@@@@@@@@@  %%%    %%%%    %%%     %%      %%%          \n"
        + "             @@@@@@  @@@  %%%     %%%%    %%%% %%         %%%%        \n"
        + "             @@@@        %%%%      %%%%%   %%%%            %%%%  %    \n"
        + "             @@          %%%%        %%%%   %%%%%@          %%%%%     \n"
        + "             @@@@        %            %%      %              %%       \n\n\n"
        + "   Explaining AI through causal Responsibility EXplanations\n"
    )


def cmdargs_parser():
    """parses command line flags"""
    parser = argparse.ArgumentParser(
        prog="ReX",
        description=f"{rex_ascii()}",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "filename",
        help="file to be processed, either an image, a numpy array, a mat file, or a nifti file",
    )
    parser.add_argument(
        "--output",
        nargs="?",
        const="show",
        help="show minimal, sufficient causal explanation, optionally saved to <OUTPUT>. Requires a PIL compatible file extension",
    )
    parser.add_argument(
        "-c", "--config", type=str, help="optional config file to use for ReX"
    )

    parser.add_argument(
        "--processed",
        action="store_true",
        help="prevent ReX from performing any preprocessing",
    )

    parser.add_argument(
        "--script",
        type=str,
        help="custom loading and preprocessing script, mostly for use with pytorch models",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=1,
        help="increase verbosity level, either -v or -vv",
    )

    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="set verbosity level to 0 (errors only), overrides --verbose",
    )

    parser.add_argument(
        "--surface",
        nargs="?",
        const="show",
        help="surface plot of the responsibility map, optionally saved to <SURFACE>",
    )
    parser.add_argument(
        "--heatmap",
        nargs="?",
        const="show",
        help="heatmap plot of the responsibility map, optionally saved to <HEATMAP>",
    )

    parser.add_argument("--model", type=str, help="model in .onnx format")

    parser.add_argument(
        "--strategy",
        "-s",
        type=str,
        help="explanation strategy, one of < multi | spatial | global | spotlight >, defaults to global",
    )
    parser.add_argument(
        "--database",
        "-db",
        type=str,
        help="store output in sqlite database <DATABASE>, creating db if necessary. Please include the extension in the filename",
    )

    parser.add_argument(
        "--multi",
        nargs="?",
        const=10,
        help="multiple explanations, with optional number <x> of spotlights, defaults to value in <rex.toml>, or 10 if undefined",
    )

    parser.add_argument(
        "--contrastive",
        nargs="?",
        const=10,
        help="a contrastive explanation, minimal, necessary and sufficient. Needs optional number <x> of floodlights, defaults to value in <rex.toml>, or 10 if undefined",
    )
    parser.add_argument(
        "--iters",
        type=int,
        help="manually override the number of iterations set in the config file",
    )

    parser.add_argument(
        "--analyze",
        action="store_true",
        help="area, entropy and (possibly) insertion/deletion curves",
    )
    parser.add_argument(
        "--analyse",
        action="store_true",
        help="area, entropy and (possibly) insertion/deletion curves",
    )

    parser.add_argument(
        "--mode",
        "-m",
        type=str,
        help="assist ReX with your input type, one of <tabular>, <spectral>, <RGB>, <voxel>, <audio>",
    )

    parser.add_argument(
        "--spectral",
        action="store_true",
        help="set ReX input type to <spectral>, shortcut for --mode spectral",
    )

    parser.add_argument("--version", action="version", version=version())

    return parser


def cmdargs():
    parser = cmdargs_parser()
    args = parser.parse_args()
    return args


def match_strategy(strategy_string):
    """gets explanation extraction strategy"""
    if strategy_string is None:
        return Strategy.Global
    if strategy_string == "multi" or strategy_string == "spotlight":
        return Strategy.MultiSpotlight
    elif strategy_string == "linear" or strategy_string == "global":
        return Strategy.Global
    elif strategy_string == "spatial":
        return Strategy.Spatial
    elif strategy_string == "contrastive":
        return Strategy.Contrastive
    else:
        logger.warning(
            "Invalid strategy '%s', reverting to default value Strategy.Global",
            strategy_string,
        )
    return Strategy.Global


def match_queue_style(qs: str) -> Queue:
    qs = qs.lower()
    if qs == "all":
        return Queue.All
    elif qs == "area":
        return Queue.Area
    elif qs == "dc":
        return Queue.DC
    elif qs == "intersection":
        return Queue.Intersection
    else:
        logger.warning(
            "Invalid queue style '%s', reverting to default value Queue.Area", qs
        )
    return Queue.Area


def shared_args(cmd_args, args: CausalArgs):
    """parses shared args"""
    if cmd_args.config is not None:
        args.config_location = cmd_args.config
    if cmd_args.model is not None:
        args.model = cmd_args.model
    if cmd_args.surface is not None:
        args.surface = cmd_args.surface
    if cmd_args.heatmap is not None:
        args.heatmap = cmd_args.heatmap
    if cmd_args.output is not None:
        args.output = cmd_args.output
    if cmd_args.quiet:
        args.verbosity = 0
    else:
        args.verbosity = cmd_args.verbose
    if cmd_args.database is not None:
        args.db = cmd_args.database
    if cmd_args.mode is not None:
        args.mode = cmd_args.mode
    if cmd_args.spectral:
        args.mode = "spectral"

    args.processed = cmd_args.processed


def find_config_path():
    conf_home = expanduser("~/.config/rex.toml")
    # search in current directory first
    if exists("rex.toml"):
        logger.info("using rex.toml in current working directory")
        config_path = "rex.toml"
    # fallback on $HOME/.config on linux/macos
    elif exists(conf_home):
        logger.info(f"using config in {conf_home}")
        config_path = conf_home
    else:
        config_path = None

    return config_path


def apply_dict_to_args(source, args, allowed_values=None):
    for k, v in source.items():
        if type(v) is not dict:
            if allowed_values is not None:
                if k not in allowed_values:
                    logger.warning("Invalid or misplaced parameter '%s', skipping!", k)
                    continue
            if hasattr(args, k):
                setattr(args, k, v)
            else:
                logger.warning(
                    "Parameter '%s' not in CausalArgs attributes, skipping!", k
                )


def validate_numeric_arg_within_bounds(n, lower, upper):
    if n < lower or n > upper:
        raise ReXTomlError(f"Invalid value '{n}': must be between {lower} and {upper}")


def validate_numeric_arg_more_than(n, lower):
    if not n > lower:
        raise ReXTomlError(f"Invalid value '{n}': must be more than {lower}")


def process_config_dict(config_file_args, args):
    expected_values = {
        "rex": ["mask_value", "seed", "gpu", "batch_size"],
        "onnx": [
            "means",
            "stds",
            "binary_threshold",
            "norm",
            "intra_op_num_threads",
            "inter_op_num_threads",
            "ort_logger",
        ],
        "visual": [
            "info",
            "colour",
            "alpha",
            "raw",
            "resize",
            "progress_bar",
            "grid",
            "mark_segments",
            "heatmap_colours",
            "multi_style",
        ],
        "causal": [
            "tree_depth",
            "search_limit",
            "iters",
            "min_box_size",
            "confidence_filter",
            "weighted",
            "queue_style",
            "queue_len",
            "concentrate",
        ],
        "distribution": ["distribution", "blend", "distribution_args"],
        "explanation": ["chunk_size", "minimum_confidence_threshold"],
        "spatial": ["spatial_initial_radius", "spatial_radius_eta", "no_expansions"],
        "multi": [
            "strategy",
            "spotlights",
            "spotlight_size",
            "spotlight_eta",
            "spotlight_step",
            "max_spotlight_budget",
            "spotlight_objective_function",
            "permitted_overlap",
        ],
        "evaluation": ["insertion_step", "normalise_curves"],
    }

    if "causal" in config_file_args.keys():
        causal_dict = config_file_args["causal"]
        apply_dict_to_args(causal_dict, args, expected_values["causal"])
        if "distribution" in causal_dict.keys():
            apply_dict_to_args(
                causal_dict["distribution"], args, expected_values["distribution"]
            )

    if "rex" in config_file_args.keys():
        rex_dict = config_file_args["rex"]
        apply_dict_to_args(rex_dict, args, expected_values["rex"])
        for subdict_name in ["visual", "onnx"]:
            if subdict_name in rex_dict.keys():
                apply_dict_to_args(
                    rex_dict[subdict_name], args, expected_values[subdict_name]
                )

    if "explanation" in config_file_args.keys():
        explain_dict = config_file_args["explanation"]
        apply_dict_to_args(explain_dict, args, expected_values["explanation"])
        for subdict_name in ["spatial", "multi", "evaluation"]:
            if subdict_name in explain_dict.keys():
                apply_dict_to_args(
                    explain_dict[subdict_name], args, expected_values[subdict_name]
                )

    if type(args.distribution) is str:
        args.distribution = str2distribution(args.distribution)
        if args.distribution == Distribution.Uniform:
            args.distribution_args = None

    if type(args.queue_style) is str:
        args.queue_style = match_queue_style(args.queue_style)

    if type(args.strategy) is str:
        args.strategy = match_strategy(args.strategy)


def process_custom_script(script, args):
    name, _ = os.path.splitext(script)
    spec = importlib.util.spec_from_file_location(name, script)
    script = importlib.util.module_from_spec(spec)  # type: ignore
    try:
        spec.loader.exec_module(script)  # type: ignore
    except Exception as e:
        logger.error("failed to load %s because of %s, exiting...", script, e)
        raise e
    args.script = script
    args.script_location = script


def process_cmd_args(cmd_args, args):
    if cmd_args.script is not None:
        try:
            process_custom_script(cmd_args.script, args)
        except Exception as e:
            logger.fatal(e)
            exit(-1)

    args.path = cmd_args.filename

    args.strategy = match_strategy(cmd_args.strategy)

    if cmd_args.iters is not None:
        args.iters = cmd_args.iters

    if cmd_args.analyze or cmd_args.analyse:
        args.analyze = True

    if cmd_args.multi is not None:
        args.strategy = Strategy.MultiSpotlight
        args.spotlights = int(cmd_args.multi)

    if cmd_args.contrastive is not None:
        args.strategy = Strategy.Contrastive
        args.spotlights = int(cmd_args.contrastive)


def load_config(config_path=None):
    if config_path is None:
        config_path = find_config_path()

    default_args = CausalArgs()
    default_args.config_location = config_path

    try:
        config_file_args = read_config_file(config_path)
        if config_file_args is None:
            logger.warning(
                "Could not find a rex.toml, so running with defaults. This might not produce the effect you want..."
            )
            return default_args
        try:
            process_config_dict(config_file_args, default_args)
            return default_args
        except Exception as e:
            logger.warning(
                "exception of type %s: %s, so reverting to default args", type(e), e
            )
            return default_args

    except ReXError as e:
        logger.fatal(e)
        exit(-1)


def get_all_args():
    """parses all arguments from config file and command line"""
    cmd_args = cmdargs()

    config_path = None
    if cmd_args.config is not None:
        config_path = cmd_args.config

    args = load_config(config_path)

    process_cmd_args(cmd_args, args)

    shared_args(cmd_args, args)

    if args.model is None and args.script_location is None:
        raise RuntimeError("either a <model>.onnx or a python file must be provided")

    return args


def validate_args(args: CausalArgs):
    """Validates a CausalArgs object.

    Checks that ``args.path`` is not None, that boolean args are boolean, and that numeric args fall within correct bounds.

    Args:
        args: configuration values for ReX
    """

    if args.path is None:
        raise FileNotFoundError("Input file path cannot be None")

    # values that must be between 0 and 1
    for arg in [
        "blend",
        "permitted_overlap",
        "alpha",
        "confidence_filter",
        "spatial_radius_eta",
        "spotlight_eta",
        "binary_threshold",
    ]:
        val = getattr(args, arg)
        if val is not None:
            validate_numeric_arg_within_bounds(val, lower=0.0, upper=1.0)

    # values that must be more than 0
    for arg in [
        "iters",
        "min_box_size",
        "chunk_size",
        "spatial_initial_radius",
        "no_expansions",
        "spotlights",
        "spotlight_size",
        "spotlight_step",
        "max_spotlight_budget",
        "insertion_step",
    ]:
        val = getattr(args, arg)
        if val is not None:
            validate_numeric_arg_more_than(val, lower=0.0)

    # values that must be boolean
    for arg in [
        "gpu",
        "info",
        "progress_bar",
        "raw",
        "resize",
        "grid",
        "mark_segments",
        "weighted",
        "concentrate",
        "normalise_curves",
    ]:
        val = getattr(args, arg)
        if val is not None:
            if not isinstance(val, bool):
                raise ReXTomlError(f"Invalid value '{val}' for {arg}, must be boolean")

    # custom treatment of specific values
    validate_numeric_arg_within_bounds(args.colour, lower=0.0, upper=255.0)

    if args.multi_style is not None:
        if args.multi_style not in ["composite", "separate"]:
            raise ReXTomlError(
                f"Invalid value '{args.multi_style}' for multi_style, must be 'composite' or 'separate'"
            )

    validate_numeric_arg_more_than(args.queue_len, lower=0.0)
    if not isinstance(args.queue_len, int):
        if args.queue_len != "all":
            raise ReXTomlError(
                f"Invalid value '{args.queue_len}' for queue_len, must be 'all' or an integer"
            )

    if args.distribution_args is not None:
        if not isinstance(args.distribution_args, list):
            raise ReXTomlError(
                f"distribution args must be a list, not '{type(args.distribution_args)}'"
            )
        elif len(args.distribution_args) != 2:
            raise ReXTomlError(
                f"distribution args must be length 2, not {len(args.distribution_args)}"
            )
        if not all([x > 0 for x in args.distribution_args]):
            raise ReXTomlError("All values in distribution args must be more than zero")

    if args.heatmap_colours not in list(mpl.colormaps):
        raise ReXTomlError(
            f"Invalid colourmap '{args.heatmap_colours}', must be a valid matplotlib colourmap"
        )



"""main entry point to ReX"""

from rex_xai._utils import get_device
from rex_xai.config import get_all_args
from rex_xai.database import initialise_rex_db

from rex_xai.explanation import explanation
from rex_xai.logger import logger, set_log_level
from rex_xai.config import validate_args


def main():
    """main entry point to ReX cmdline tool"""
    args = get_all_args()
    validate_args(args)
    set_log_level(args.verbosity, logger)

    device = get_device(args.gpu)

    logger.debug("running ReX with the following args:\n %s", args)

    db = None
    if args.db is not None:
        db = initialise_rex_db(args.db)

    explanation(args, device, db)



#!/usr/bin/env python
from typing import Tuple
import numpy as np
import torch as tt
from scipy.integrate import simpson
from scipy.signal import periodogram
from skimage.measure import shannon_entropy

from rex_xai.extraction import Explanation
from rex_xai._utils import get_map_locations
from rex_xai.mutant import _apply_to_data
from rex_xai._utils import set_boolean_mask_value, xlogx


class Evaluation:
    # TODO does this need to be an object? Probably not...
    # TODO consider inheritance from Explanation object
    def __init__(self, explanation: Explanation) -> None:
        self.explanation = explanation

    def ratio(self) -> float:
        """Returns percentage of data required for sufficient explanation"""
        if (
            self.explanation.explanation is None
            or self.explanation.data.model_channels is None
        ):
            raise ValueError("Invalid Explanation object")

        try:
            final_mask = self.explanation.final_mask.squeeze().item()  # type: ignore
        except Exception:
            final_mask = self.explanation.final_mask

        try:
            return (
                tt.count_nonzero(final_mask)  # type: ignore
                / final_mask.size  # type: ignore
                / self.explanation.data.model_channels
            ).item()
        except TypeError:
            return (
                np.count_nonzero(final_mask)  # type: ignore
                / final_mask.size  # type: ignore
                / self.explanation.data.model_channels
            )

    def spectral_entropy(self) -> Tuple[float, float]:
        """
        This code is a simplified version of
        https://github.com/raphaelvallat/antropy/blob/master/src/antropy/entropy.py
        """
        _, psd = periodogram(self.explanation.target_map)
        psd_norm = psd / psd.sum()
        ent = -np.sum(xlogx(psd_norm))
        if len(psd_norm.shape) == 2:
            max_ent = np.log2(len(psd_norm[0]))
        else:
            max_ent = np.log2(len(psd_norm))
        return ent, max_ent

    def entropy_loss(self):
        img = np.array(self.explanation.data.input)
        assert self.explanation.explanation is not None
        exp = shannon_entropy(self.explanation.explanation.detach().cpu().numpy())

        return shannon_entropy(img), exp

    def insertion_deletion_curve(self, prediction_func, normalise=False):
        insertion_curve = []
        deletion_curve = []

        assert self.explanation.data.target is not None
        assert self.explanation.data.target.confidence is not None

        assert self.explanation.data.data is not None
        insertion_mask = tt.zeros(
            self.explanation.data.data.squeeze(0).shape, dtype=tt.bool
        ).to(self.explanation.data.device)
        deletion_mask = tt.ones(
            self.explanation.data.data.squeeze(0).shape, dtype=tt.bool
        ).to(self.explanation.data.device)
        im = []
        dm = []

        step = self.explanation.args.insertion_step
        ranking = get_map_locations(map=self.explanation.target_map)
        iters = len(ranking) // step

        for i in range(0, len(ranking), step):
            chunk = ranking[i : i + step]
            for _, loc in chunk:
                set_boolean_mask_value(
                    insertion_mask,
                    self.explanation.data.mode,
                    self.explanation.data.model_order,
                    loc,
                )
                set_boolean_mask_value(
                    deletion_mask,
                    self.explanation.data.mode,
                    self.explanation.data.model_order,
                    loc,
                    val=False,
                )
            im.append(
                _apply_to_data(insertion_mask, self.explanation.data, 0).squeeze(0)
            )
            dm.append(
                _apply_to_data(deletion_mask, self.explanation.data, 0).squeeze(0)
            )

            if len(im) == self.explanation.args.batch_size:
                self.__batch(im, dm, prediction_func, insertion_curve, deletion_curve)
                im = []
                dm = []

        if im != [] and dm != []:
            self.__batch(im, dm, prediction_func, insertion_curve, deletion_curve)

        i_auc = simpson(insertion_curve, dx=step)
        d_auc = simpson(deletion_curve, dx=step)

        if normalise:
            const = self.explanation.data.target.confidence * iters * step
            i_auc /= const
            d_auc /= const

        return i_auc, d_auc

    # # def sensitivity(self):
    #     #     pass

    # # def infidelity(self):
    #     #     pass

    def __batch(self, im, dm, prediction_func, insertion_curve, deletion_curve):
        assert self.explanation.data.target is not None
        ip = prediction_func(tt.stack(im).to(self.explanation.data.device), raw=True)
        dp = prediction_func(tt.stack(dm).to(self.explanation.data.device), raw=True)
        for p in range(0, ip.shape[0]):
            insertion_curve.append(
                ip[p, self.explanation.data.target.classification].item()
            )  # type: ignore
            deletion_curve.append(
                dp[p, self.explanation.data.target.classification].item()
            )  # type: ignore
